import numba as nb
import numpy as np

from oasislmf.pytools.common.data import DEFAULT_BUFFER_SIZE, nb_oasis_int
from oasislmf.pytools.lec.data import (EPT_dtype, PSEPT_dtype, TAIL_valtype, NB_TAIL_valtype)
from oasislmf.pytools.lec.utils import (create_empty_array, get_wheatsheaf_items_idx, get_wheatsheaf_items_idx_data, resize_array)


@nb.njit(cache=True, error_model="numpy")
def get_loss(
    next_retperiod,
    last_retperiod,
    last_loss,
    curr_retperiod,
    curr_loss
):
    """Get loss based on current and next return period
    Args:
        next_retperiod (float): Next return period
        last_retperiod (float): Previous return period
        last_loss (float): Previous Loss value
        curr_retperiod (float): Current return period
        curr_loss (float): Current Loss value
    Returns:
        loss (float): Loss Value
    """
    if curr_retperiod == 0:
        return 0
    if curr_loss == 0:
        return 0
    if curr_retperiod == next_retperiod:
        return curr_loss
    if curr_retperiod < next_retperiod:
        # Linear Interpolate
        return (
            (next_retperiod - curr_retperiod) * (last_loss - curr_loss) /
            (last_retperiod - curr_retperiod) + curr_loss
        )
    return -1


@nb.njit(cache=True, error_model="numpy")
def fill_tvar(
    tail,
    tail_sizes,
    summary_id,
    next_retperiod,
    tvar
):
    """Populate the Tail with retperiod and tvar values for summary_id
    Args:
        tail (nb.typed.Dict[nb_oasis_int, NB_TAIL_valtype]): Dict of Summary ID to vector of (return period, tvar) values
        tail_sizes (nb.typed.Dict[nb_oasis_int, nb.types.int64]): Dict of Summary ID to size of each tail array
        summary_id (int): Summary ID
        next_retperiod (float): Next Return Period
        tvar (float): Tail Value at Risk
    Returns:
        tail (nb.typed.Dict[nb_oasis_int, NB_TAIL_valtype]): Dict of summary_id to vector of (return period, tvar) values
        tail_sizes (nb.typed.Dict[nb_oasis_int, nb.types.int64]): Dict of summary_id to size of each tail array
    """
    if summary_id not in tail:
        tail[summary_id] = create_empty_array(TAIL_valtype)
        tail_sizes[summary_id] = 0
    tail_arr = tail[summary_id]
    tail_arr = resize_array(tail_arr, tail_sizes[summary_id])
    tail_current_size = tail_sizes[summary_id]
    tail_arr[tail_current_size]["retperiod"] = next_retperiod
    tail_arr[tail_current_size]["tvar"] = tvar
    tail[summary_id] = tail_arr
    tail_sizes[summary_id] += 1

    return tail, tail_sizes


@nb.njit(cache=True, error_model="numpy")
def fill_tvar_wheatsheaf(
    tail,
    tail_sizes,
    summary_id,
    sidx,
    num_sidxs,
    next_retperiod,
    tvar
):
    """Populate the Tail with retperiod and tvar values for (summary_id, sidx) pair
    Args:
        tail (nb.typed.Dict[nb_oasis_int, NB_TAIL_valtype]): Dict of (summary_id, sidx) pair to vector of (return period, tvar) values
        tail_sizes (nb.typed.Dict[nb_oasis_int, nb.types.int64]): Dict of (summary_id, sidx) pair to size of each tail array
        summary_id (int): Summary ID
        sidx (int): Sample ID
        num_sidxs (int): Number of sidxs to consider
        next_retperiod (float): Next Return Period
        tvar (float): Tail Value at Risk
    Returns:
        tail (nb.typed.Dict[nb_oasis_int, NB_TAIL_valtype]): Dict of (summary_id, sidx) pair to vector of (return period, tvar) values
        tail_sizes (nb.typed.Dict[nb_oasis_int, nb.types.int64]): Dict of (summary_id, sidx) pair to size of each tail array
    """
    idx = get_wheatsheaf_items_idx(summary_id, sidx, num_sidxs)
    if idx not in tail:
        tail[idx] = create_empty_array(TAIL_valtype)
        tail_sizes[idx] = 0
    tail_arr = tail[idx]
    tail_arr = resize_array(tail_arr, tail_sizes[idx])
    tail_current_size = tail_sizes[idx]
    tail_arr[tail_current_size]["retperiod"] = next_retperiod
    tail_arr[tail_current_size]["tvar"] = tvar
    tail[idx] = tail_arr
    tail_sizes[idx] += 1

    return tail, tail_sizes


@nb.njit(cache=True, error_model="numpy")
def write_return_period_out(
    next_returnperiod_idx,
    last_computed_rp,
    last_computed_loss,
    curr_retperiod,
    curr_loss,
    summary_id,
    eptype,
    epcalc,
    max_retperiod,
    counter,
    tvar,
    tail,
    tail_sizes,
    returnperiods,
    mean_map=None,
    is_wheatsheaf=False,
    num_sidxs=-1,
):
    """Processes return periods and computes losses for a given summary, updating TVaR and mean map if required.
    Args:
        next_returnperiod_idx (int): Index of the next return period to process.
        last_computed_rp (float): Last computed return period
        last_computed_loss (float): Last computed loss value
        curr_retperiod (float): Current return period being processed.
        curr_loss (float): Loss associated with the current return period.
        summary_id (int): Identifier for the current summary.
        eptype (int): Type of exceedance probability (0 = OEP, 1 = AEP).
        epcalc (int): Type of exceedance probability calculation.
        max_retperiod (int): Maximum return period to be used in calculations
        counter (int): Counter used for updating TVaR
        tvar (float): Tail Value at Risk
        tail (nb.typed.Dict[nb_oasis_int, NB_TAIL_valtype]): Dict of summary_id or (summary_id, sidx) pair to vector of (return period, tvar) values
        tail_sizes (nb.typed.Dict[nb_oasis_int, nb.types.int64]): Dict of summary_id or (summary_id, sidx) pair to size of each tail array
        returnperiods (ndarray[np.int32]): Return Periods array
        mean_map (ndarray[MEANMAP_dtype], optional): An array mapping used for mean loss calculations per Summary ID. Used for EPT output later. Defaults to None.
        is_wheatsheaf (bool, optional): If True, update the wheatsheaf TVaR structure.
        num_sidxs (int, optional): Number of sidxs to consider. Defaults to -1 if not is_wheatsheaf.
    Returns:
        rets (list[EPT_dtype]): Return period and Loss EPT data
        tail (nb.typed.Dict[nb_oasis_int, NB_TAIL_valtype]): Dict of summary_id or (summary_id, sidx) pair to vector of (return period, tvar) values
        tail_sizes (nb.typed.Dict[nb_oasis_int, nb.types.int64]): Dict of summary_id or (summary_id, sidx) pair to size of each tail array
        last_computed_rp (float): Last computed return period
        last_computed_loss (float): Last computed loss value
    """
    next_retperiod = 0
    rets = []
    while True:
        if next_returnperiod_idx >= len(returnperiods):
            return rets, tail, tail_sizes, next_returnperiod_idx, last_computed_rp, last_computed_loss

        next_retperiod = returnperiods[next_returnperiod_idx]

        if curr_retperiod > next_retperiod:
            break

        if max_retperiod < next_retperiod:
            next_returnperiod_idx += 1
            continue

        loss = get_loss(
            next_retperiod,
            last_computed_rp,
            last_computed_loss,
            curr_retperiod,
            curr_loss
        )

        rets.append((summary_id, epcalc, eptype, next_retperiod, loss))

        if mean_map is not None:
            mean_map[summary_id - 1][next_returnperiod_idx]["retperiod"] = next_retperiod
            mean_map[summary_id - 1][next_returnperiod_idx]["mean"] += loss
            mean_map[summary_id - 1][next_returnperiod_idx]["count"] += 1

        if curr_retperiod != 0:
            tvar = tvar - ((tvar - loss) / counter)
            if is_wheatsheaf:
                tail, tail_sizes = fill_tvar_wheatsheaf(
                    tail,
                    tail_sizes,
                    summary_id,
                    epcalc,
                    num_sidxs,
                    next_retperiod,
                    tvar,
                )
            else:
                tail, tail_sizes = fill_tvar(
                    tail,
                    tail_sizes,
                    summary_id,
                    next_retperiod,
                    tvar,
                )

        next_returnperiod_idx += 1
        counter += 1

        if curr_retperiod > next_retperiod:
            break

    if curr_retperiod > 0:
        last_computed_rp = curr_retperiod
        last_computed_loss = curr_loss
    return rets, tail, tail_sizes, next_returnperiod_idx, last_computed_rp, last_computed_loss


@nb.njit(cache=True, error_model="numpy")
def write_tvar(
    epcalc,
    eptype_tvar,
    tail,
    tail_sizes,
):
    """Get TVaR values for EPT output from tail
    Args:
        epcalc (int): Type of exceedance probability calculation.
        eptype_tvar (int): Type of Tail Value-at-Risk (TVAR) to calculate (0 = OEP TVAR, 1 = AEP TVAR).
        tail (nb.typed.Dict[nb_oasis_int, NB_TAIL_valtype]): Dict of summary_id to vector of (return period, tvar) values
        tail_sizes (nb.typed.Dict[nb_oasis_int, nb.types.int64]): Dict of summary_id pair to size of each tail array
    Returns:
        rets (list[EPT_dtype]): Return period and Loss EPT data
    """
    rets = []
    for summary_id in sorted(tail.keys()):
        vals = tail[summary_id][:tail_sizes[summary_id]]
        for row in vals:
            rets.append((summary_id, epcalc, eptype_tvar, row["retperiod"], row["tvar"]))
    return rets


@nb.njit(cache=True, error_model="numpy")
def write_tvar_wheatsheaf(
    num_sidxs,
    eptype_tvar,
    tail,
    tail_sizes,
):
    """Get TVaR values for PSEPT output from tail
    Args:
        num_sidxs (int): Number of sidxs to consider.
        eptype_tvar (int): Type of Tail Value-at-Risk (TVAR) to calculate (0 = OEP TVAR, 1 = AEP TVAR).
        tail (nb.typed.Dict[nb_oasis_int, NB_TAIL_valtype]): Dict of (summary_id, sidx) pair to vector of (return period, tvar) values
        tail_sizes (nb.typed.Dict[nb_oasis_int, nb.types.int64]): Dict of (summary_id, sidx) pair to size of each tail array
    Returns:
        rets (list[PSEPT_dtype]): Return period and Loss PSEPT data
    """
    rets = []
    for idx in sorted(tail.keys()):
        sidx, summary_id = get_wheatsheaf_items_idx_data(idx, num_sidxs)
        vals = tail[idx][:tail_sizes[idx]]
        for row in vals:
            rets.append((summary_id, sidx, eptype_tvar, row["retperiod"], row["tvar"]))
    return rets


@nb.njit(cache=True, error_model="numpy")
def write_ept(
    items,
    items_start_end,
    max_retperiod,
    epcalc,
    eptype,
    eptype_tvar,
    use_return_period,
    returnperiods,
    max_summary_id,
    sample_size=1
):
    """Generate Loss Exceedance Curve values and Tail Value at Risk values based on items and epcalc/eptype/eptype_tvar

    The loss calculation follows these principles:
    - For Aggregate Loss Exceedance Curves (AEP): The sum of all losses within a period is calculated.
    - For Occurrence Loss Exceedance Curves (OEP): The maximum loss within a period is taken.
    - TVAR (Tail Conditional Expectation): Calculated as the average of losses exceeding a given return period.
    Args:
        items (ndarray[LOSSVEC2MAP_dtype]): Array mapping summary_id to loss value (and period_no/period_weighting where applicable)
        items_start_end (ndarray[np.int32]): An array marking where the start and end idxs are for each summary_id in the items array 
        max_retperiod (int): Maximum return period to be used in calculations
        epcalc (int): Specifies the calculation method (mean damage loss, full uncertainty, per sample mean, sample mean).
        eptype (int): Type of exceedance probability (0 = OEP, 1 = AEP).
        eptype_tvar (int): Type of Tail Value-at-Risk (TVAR) to calculate (0 = OEP TVAR, 1 = AEP TVAR).
        use_return_period (bool): Use Return Period file.
        returnperiods (ndarray[np.int32]): Return Periods array
        max_summary_id (int): Maximum summary ID
        sample_size (int, optional): Sample Size. Defaults to 1.
    Yields:
        buffer (ndarray[EPT_dtype]): Buffered chunks of EPT data
    """
    buffer = np.zeros(DEFAULT_BUFFER_SIZE, dtype=EPT_dtype)
    bidx = 0

    if len(items) == 0 or sample_size == 0:
        return

    tail = nb.typed.Dict.empty(nb_oasis_int, NB_TAIL_valtype)
    tail_sizes = nb.typed.Dict.empty(nb_oasis_int, nb.types.int64)

    for summary_id in range(1, max_summary_id + 1):
        start, end = items_start_end[summary_id - 1]
        if start == -1:
            continue
        filtered_items = items[start:end]
        sorted_idxs = np.argsort(filtered_items["value"])[::-1]
        sorted_items = filtered_items[sorted_idxs]
        next_returnperiod_idx = 0
        last_computed_rp = 0
        last_computed_loss = 0
        tvar = 0
        i = 1
        for item in sorted_items:
            value = item["value"] / sample_size
            retperiod = max_retperiod / i

            if use_return_period:
                if next_returnperiod_idx < len(returnperiods):
                    rets, tail, tail_sizes, next_returnperiod_idx, last_computed_rp, last_computed_loss = write_return_period_out(
                        next_returnperiod_idx,
                        last_computed_rp,
                        last_computed_loss,
                        retperiod,
                        value,
                        summary_id,
                        eptype,
                        epcalc,
                        max_retperiod,
                        i,
                        tvar,
                        tail,
                        tail_sizes,
                        returnperiods,
                    )
                    for ret in rets:
                        if bidx >= len(buffer):
                            yield buffer[:bidx]
                            bidx = 0
                        buffer[bidx]["SummaryId"] = ret[0]
                        buffer[bidx]["EPCalc"] = ret[1]
                        buffer[bidx]["EPType"] = ret[2]
                        buffer[bidx]["ReturnPeriod"] = ret[3]
                        buffer[bidx]["Loss"] = ret[4]
                        bidx += 1
                tvar = tvar - ((tvar - value) / i)
            else:
                tvar = tvar - ((tvar - value) / i)

                if summary_id not in tail:
                    tail[summary_id] = create_empty_array(TAIL_valtype)
                    tail_sizes[summary_id] = 0
                tail_arr = tail[summary_id]
                tail_arr = resize_array(tail_arr, tail_sizes[summary_id])
                tail_current_size = tail_sizes[summary_id]
                tail_arr[tail_current_size]["retperiod"] = retperiod
                tail_arr[tail_current_size]["tvar"] = tvar
                tail[summary_id] = tail_arr
                tail_sizes[summary_id] += 1

                if bidx >= len(buffer):
                    yield buffer[:bidx]
                    bidx = 0
                buffer[bidx]["SummaryId"] = summary_id
                buffer[bidx]["EPCalc"] = epcalc
                buffer[bidx]["EPType"] = eptype
                buffer[bidx]["ReturnPeriod"] = retperiod
                buffer[bidx]["Loss"] = value
                bidx += 1

            i += 1

        if use_return_period:
            while True:
                retperiod = max_retperiod / i
                if returnperiods[next_returnperiod_idx] <= 0:
                    retperiod = 0
                if next_returnperiod_idx < len(returnperiods):
                    rets, tail, tail_sizes, next_returnperiod_idx, last_computed_rp, last_computed_loss = write_return_period_out(
                        next_returnperiod_idx,
                        last_computed_rp,
                        last_computed_loss,
                        retperiod,
                        0,
                        summary_id,
                        eptype,
                        epcalc,
                        max_retperiod,
                        i,
                        tvar,
                        tail,
                        tail_sizes,
                        returnperiods,
                    )
                    for ret in rets:
                        if bidx >= len(buffer):
                            yield buffer[:bidx]
                            bidx = 0
                        buffer[bidx]["SummaryId"] = ret[0]
                        buffer[bidx]["EPCalc"] = ret[1]
                        buffer[bidx]["EPType"] = ret[2]
                        buffer[bidx]["ReturnPeriod"] = ret[3]
                        buffer[bidx]["Loss"] = ret[4]
                        bidx += 1
                tvar = tvar - (tvar / i)
                i += 1
                if next_returnperiod_idx >= len(returnperiods):
                    break

    rets = write_tvar(
        epcalc,
        eptype_tvar,
        tail,
        tail_sizes,
    )
    for ret in rets:
        if bidx >= len(buffer):
            yield buffer[:bidx]
            bidx = 0
        buffer[bidx]["SummaryId"] = ret[0]
        buffer[bidx]["EPCalc"] = ret[1]
        buffer[bidx]["EPType"] = ret[2]
        buffer[bidx]["ReturnPeriod"] = ret[3]
        buffer[bidx]["Loss"] = ret[4]
        bidx += 1
    yield buffer[:bidx]


@nb.njit(cache=True, error_model="numpy")
def write_ept_weighted(
    items,
    items_start_end,
    cum_weight_constant,
    epcalc,
    eptype,
    eptype_tvar,
    unused_period_weights,
    use_return_period,
    returnperiods,
    max_summary_id,
    sample_size=1
):
    """Generate Loss Exceedance Curve values and Tail Value at Risk values based on items and epcalc/eptype/eptype_tvar.

    This function calculates weighted exceedance probability tables using cumulative period weightings (`period_weighting`), 
    which impact the calculation of return periods. The weighting allows for more accurate representation of losses when 
    event periods have different probabilities or frequencies of occurrence.

    The loss calculation follows these principles:
    - For Aggregate Loss Exceedance Curves (AEP): The sum of all losses within a period is calculated.
    - For Occurrence Loss Exceedance Curves (OEP): The maximum loss within a period is taken.
    - TVAR (Tail Conditional Expectation): Calculated as the average of losses exceeding a given return period.
    Args:
        items (ndarray[LOSSVEC2MAP_dtype]): Array mapping summary_id to loss value (and period_no/period_weighting where applicable)
        items_start_end (ndarray[np.int32]): An array marking where the start and end idxs are for each summary_id in the items array 
        cum_weight_constant (float): Constant factor for scaling cumulative period weights.
        epcalc (int): Specifies the calculation method (mean damage loss, full uncertainty, per sample mean, sample mean).
        eptype (int): Type of exceedance probability (0 = OEP, 1 = AEP).
        eptype_tvar (int): Type of Tail Value-at-Risk (TVAR) to calculate (0 = OEP TVAR, 1 = AEP TVAR).
        unused_period_weights (ndarray[float]): Array of unused period weights
        use_return_period (bool): Use Return Period file.
        returnperiods (ndarray[np.int32]): Return Periods array
        max_summary_id (int): Maximum summary ID
        sample_size (int, optional): Sample Size. Defaults to 1.
    Yields:
        buffer (ndarray[EPT_dtype]): Buffered chunks of EPT data
    """
    buffer = np.zeros(DEFAULT_BUFFER_SIZE, dtype=EPT_dtype)
    bidx = 0

    if len(items) == 0 or sample_size == 0:
        return

    tail = nb.typed.Dict.empty(nb_oasis_int, NB_TAIL_valtype)
    tail_sizes = nb.typed.Dict.empty(nb_oasis_int, nb.types.int64)

    for summary_id in range(1, max_summary_id + 1):
        start, end = items_start_end[summary_id - 1]
        if start == -1:
            continue
        filtered_items = items[start:end]
        sorted_idxs = np.argsort(filtered_items["value"])[::-1]
        sorted_items = filtered_items[sorted_idxs]
        next_returnperiod_idx = 0
        last_computed_rp = 0
        last_computed_loss = 0
        tvar = 0
        i = 1
        cumulative_weighting = 0
        max_retperiod = 0
        largest_loss = False

        for item in sorted_items:
            value = item["value"] / sample_size
            cumulative_weighting += (item["period_weighting"] * cum_weight_constant)
            retperiod = max_retperiod / i

            if item["period_weighting"]:
                retperiod = 1 / cumulative_weighting

                if not largest_loss:
                    max_retperiod = retperiod + 0.0001  # Add for floating point errors
                    largest_loss = True

                if use_return_period:
                    if next_returnperiod_idx < len(returnperiods):
                        rets, tail, tail_sizes, next_returnperiod_idx, last_computed_rp, last_computed_loss = write_return_period_out(
                            next_returnperiod_idx,
                            last_computed_rp,
                            last_computed_loss,
                            retperiod,
                            value,
                            summary_id,
                            eptype,
                            epcalc,
                            max_retperiod,
                            i,
                            tvar,
                            tail,
                            tail_sizes,
                            returnperiods,
                        )
                        for ret in rets:
                            if bidx >= len(buffer):
                                yield buffer[:bidx]
                                bidx = 0
                            buffer[bidx]["SummaryId"] = ret[0]
                            buffer[bidx]["EPCalc"] = ret[1]
                            buffer[bidx]["EPType"] = ret[2]
                            buffer[bidx]["ReturnPeriod"] = ret[3]
                            buffer[bidx]["Loss"] = ret[4]
                            bidx += 1
                    tvar = tvar - ((tvar - (value)) / i)
                else:
                    tvar = tvar - ((tvar - (value)) / i)

                    if summary_id not in tail:
                        tail[summary_id] = create_empty_array(TAIL_valtype)
                        tail_sizes[summary_id] = 0
                    tail_arr = tail[summary_id]
                    tail_arr = resize_array(tail_arr, tail_sizes[summary_id])
                    tail_current_size = tail_sizes[summary_id]
                    tail_arr[tail_current_size]["retperiod"] = retperiod
                    tail_arr[tail_current_size]["tvar"] = tvar
                    tail[summary_id] = tail_arr
                    tail_sizes[summary_id] += 1

                    if bidx >= len(buffer):
                        yield buffer[:bidx]
                        bidx = 0
                    buffer[bidx]["SummaryId"] = summary_id
                    buffer[bidx]["EPCalc"] = epcalc
                    buffer[bidx]["EPType"] = eptype
                    buffer[bidx]["ReturnPeriod"] = retperiod
                    buffer[bidx]["Loss"] = value
                    bidx += 1

                i += 1
        if use_return_period:
            unused_pw_idx = 0
            while True:
                retperiod = 0
                if unused_pw_idx < len(unused_period_weights):
                    cumulative_weighting += (
                        unused_period_weights[unused_pw_idx]["weighting"] * cum_weight_constant
                    )
                    retperiod = 1 / cumulative_weighting
                    unused_pw_idx += 1
                if next_returnperiod_idx < len(returnperiods):
                    rets, tail, tail_sizes, next_returnperiod_idx, last_computed_rp, last_computed_loss = write_return_period_out(
                        next_returnperiod_idx,
                        last_computed_rp,
                        last_computed_loss,
                        retperiod,
                        0,
                        summary_id,
                        eptype,
                        epcalc,
                        max_retperiod,
                        i,
                        tvar,
                        tail,
                        tail_sizes,
                        returnperiods,
                    )
                    for ret in rets:
                        if bidx >= len(buffer):
                            yield buffer[:bidx]
                            bidx = 0
                        buffer[bidx]["SummaryId"] = ret[0]
                        buffer[bidx]["EPCalc"] = ret[1]
                        buffer[bidx]["EPType"] = ret[2]
                        buffer[bidx]["ReturnPeriod"] = ret[3]
                        buffer[bidx]["Loss"] = ret[4]
                        bidx += 1
                tvar = tvar - (tvar / i)
                i += 1
                if next_returnperiod_idx >= len(returnperiods):
                    break

    rets = write_tvar(
        epcalc,
        eptype_tvar,
        tail,
        tail_sizes,
    )
    for ret in rets:
        if bidx >= len(buffer):
            yield buffer[:bidx]
            bidx = 0
        buffer[bidx]["SummaryId"] = ret[0]
        buffer[bidx]["EPCalc"] = ret[1]
        buffer[bidx]["EPType"] = ret[2]
        buffer[bidx]["ReturnPeriod"] = ret[3]
        buffer[bidx]["Loss"] = ret[4]
        bidx += 1
    yield buffer[:bidx]


@nb.njit(cache=True, error_model="numpy")
def write_psept(
    items,
    items_start_end,
    max_retperiod,
    eptype,
    eptype_tvar,
    use_return_period,
    returnperiods,
    max_summary_id,
    num_sidxs
):
    """Generate Per Sample Exceedance Probability Tables (PSEPT) for each individual sample, producing a separate loss
    exceedance curve for each sample, eptype, eptype_tvar.
    Args:
        items (ndarray[WHEATKEYITEMS_dtype]): Array mapping (summary_id, sidx) to loss value (and period_no/period_weighting where applicable)
        items_start_end (ndarray[np.int32]): An array marking where the start and end idxs are for each (summary_id, sidx) pair in the items array 
        max_retperiod (int): Maximum return period to be used in calculations
        eptype (int): Type of exceedance probability (0 = OEP, 1 = AEP).
        eptype_tvar (int): Type of Tail Value-at-Risk (TVAR) to calculate (0 = OEP TVAR, 1 = AEP TVAR).
        use_return_period (bool): Use Return Period file.
        returnperiods (ndarray[np.int32]): Return Periods array
        max_summary_id (int): Maximum summary ID
        num_sidxs (int): Number of sidxs to consider
    Yields:
        buffer (ndarray[PSEPT_dtype]): Buffered chunks of PSEPT data
    """
    buffer = np.zeros(DEFAULT_BUFFER_SIZE, dtype=PSEPT_dtype)
    bidx = 0

    if len(items) == 0:
        return

    tail = nb.typed.Dict.empty(nb_oasis_int, NB_TAIL_valtype)
    tail_sizes = nb.typed.Dict.empty(nb_oasis_int, nb.types.int64)

    for idx in range(max_summary_id * num_sidxs):
        start, end = items_start_end[idx]
        if start == -1:
            continue
        sidx, summary_id = get_wheatsheaf_items_idx_data(idx, num_sidxs)
        filtered_items = items[start:end]
        sorted_idxs = np.argsort(filtered_items["value"])[::-1]
        sorted_items = filtered_items[sorted_idxs]
        next_returnperiod_idx = 0
        last_computed_rp = 0
        last_computed_loss = 0
        tvar = 0
        i = 1
        for item in sorted_items:
            value = item["value"]
            retperiod = max_retperiod / i

            if use_return_period:
                if next_returnperiod_idx < len(returnperiods):
                    rets, tail, tail_sizes, next_returnperiod_idx, last_computed_rp, last_computed_loss = write_return_period_out(
                        next_returnperiod_idx,
                        last_computed_rp,
                        last_computed_loss,
                        retperiod,
                        value,
                        summary_id,
                        eptype,
                        sidx,
                        max_retperiod,
                        i,
                        tvar,
                        tail,
                        tail_sizes,
                        returnperiods,
                        is_wheatsheaf=True,
                        num_sidxs=num_sidxs,
                    )
                    for ret in rets:
                        if bidx >= len(buffer):
                            yield buffer[:bidx]
                            bidx = 0
                        buffer[bidx]["SummaryId"] = ret[0]
                        buffer[bidx]["SampleId"] = ret[1]
                        buffer[bidx]["EPType"] = ret[2]
                        buffer[bidx]["ReturnPeriod"] = ret[3]
                        buffer[bidx]["Loss"] = ret[4]
                        bidx += 1
                tvar = tvar - ((tvar - value) / i)
            else:
                tvar = tvar - ((tvar - value) / i)

                if idx not in tail:
                    tail[idx] = create_empty_array(TAIL_valtype)
                    tail_sizes[idx] = 0
                tail_arr = tail[idx]
                tail_arr = resize_array(tail_arr, tail_sizes[idx])
                tail_current_size = tail_sizes[idx]
                tail_arr[tail_current_size]["retperiod"] = retperiod
                tail_arr[tail_current_size]["tvar"] = tvar
                tail[idx] = tail_arr
                tail_sizes[idx] += 1

                if bidx >= len(buffer):
                    yield buffer[:bidx]
                    bidx = 0
                buffer[bidx]["SummaryId"] = summary_id
                buffer[bidx]["SampleId"] = sidx
                buffer[bidx]["EPType"] = eptype
                buffer[bidx]["ReturnPeriod"] = retperiod
                buffer[bidx]["Loss"] = value
                bidx += 1

            i += 1

        if use_return_period:
            while True:
                retperiod = max_retperiod / i
                if returnperiods[next_returnperiod_idx] <= 0:
                    retperiod = 0
                if next_returnperiod_idx < len(returnperiods):
                    rets, tail, tail_sizes, next_returnperiod_idx, last_computed_rp, last_computed_loss = write_return_period_out(
                        next_returnperiod_idx,
                        last_computed_rp,
                        last_computed_loss,
                        retperiod,
                        0,
                        summary_id,
                        eptype,
                        sidx,
                        max_retperiod,
                        i,
                        tvar,
                        tail,
                        tail_sizes,
                        returnperiods,
                        is_wheatsheaf=True,
                        num_sidxs=num_sidxs,
                    )
                    for ret in rets:
                        if bidx >= len(buffer):
                            yield buffer[:bidx]
                            bidx = 0
                        buffer[bidx]["SummaryId"] = ret[0]
                        buffer[bidx]["SampleId"] = ret[1]
                        buffer[bidx]["EPType"] = ret[2]
                        buffer[bidx]["ReturnPeriod"] = ret[3]
                        buffer[bidx]["Loss"] = ret[4]
                        bidx += 1
                tvar = tvar - (tvar / i)
                i += 1
                if next_returnperiod_idx >= len(returnperiods):
                    break

    rets = write_tvar_wheatsheaf(
        num_sidxs,
        eptype_tvar,
        tail,
        tail_sizes,
    )
    for ret in rets:
        if bidx >= len(buffer):
            yield buffer[:bidx]
            bidx = 0
        buffer[bidx]["SummaryId"] = ret[0]
        buffer[bidx]["SampleId"] = ret[1]
        buffer[bidx]["EPType"] = ret[2]
        buffer[bidx]["ReturnPeriod"] = ret[3]
        buffer[bidx]["Loss"] = ret[4]
        bidx += 1
    yield buffer[:bidx]


@nb.njit(cache=True, error_model="numpy")
def write_psept_weighted(
    items,
    items_start_end,
    max_retperiod,
    eptype,
    eptype_tvar,
    unused_period_weights,
    use_return_period,
    returnperiods,
    max_summary_id,
    num_sidxs,
    sample_size,
    mean_map=None
):
    """Generate Per Sample Exceedance Probability Tables (PSEPT) for each individual sample, producing a separate loss
    exceedance curve for each sample, eptype, eptype_tvar.
    Args:
        items (ndarray[WHEATKEYITEMS_dtype]): Array mapping (summary_id, sidx) to loss value (and period_no/period_weighting where applicable)
        items_start_end (ndarray[np.int32]): An array marking where the start and end idxs are for each (summary_id, sidx) pair in the items array 
        max_retperiod (int): Maximum return period to be used in calculations
        eptype (int): Type of exceedance probability (0 = OEP, 1 = AEP).
        eptype_tvar (int): Type of Tail Value-at-Risk (TVAR) to calculate (0 = OEP TVAR, 1 = AEP TVAR).
        unused_period_weights (ndarray[float]): Array of unused period weights
        use_return_period (bool): Use Return Period file.
        returnperiods (ndarray[np.int32]): Return Periods array
        max_summary_id (int): Maximum summary ID
        num_sidxs (int): Number of sidxs to consider
        sample_size (int): Sample Size. Defaults to 1.
        mean_map (ndarray[MEANMAP_dtype], optional): An array mapping used for mean loss calculations per Summary ID. Used for EPT output later. Defaults to None.
    Yields:
        buffer (ndarray[PSEPT_dtype]): Buffered chunks of PSEPT data
    """
    buffer = np.zeros(DEFAULT_BUFFER_SIZE, dtype=PSEPT_dtype)
    bidx = 0

    if len(items) == 0:
        return

    tail = nb.typed.Dict.empty(nb_oasis_int, NB_TAIL_valtype)
    tail_sizes = nb.typed.Dict.empty(nb_oasis_int, nb.types.int64)

    for idx in range(max_summary_id * num_sidxs):
        start, end = items_start_end[idx]
        if start == -1:
            continue
        sidx, summary_id = get_wheatsheaf_items_idx_data(idx, num_sidxs)
        filtered_items = items[start:end]
        sorted_idxs = np.argsort(filtered_items["value"])[::-1]
        sorted_items = filtered_items[sorted_idxs]
        next_returnperiod_idx = 0
        last_computed_rp = 0
        last_computed_loss = 0
        tvar = 0
        i = 1
        cumulative_weighting = 0
        max_retperiod = 0
        largest_loss = False

        for item in sorted_items:
            value = item["value"]
            cumulative_weighting += (item["period_weighting"] * sample_size)
            retperiod = max_retperiod / i

            if item["period_weighting"]:
                retperiod = 1 / cumulative_weighting

                if not largest_loss:
                    max_retperiod = retperiod + 0.0001  # Add for floating point errors
                    largest_loss = True

                if use_return_period:
                    if next_returnperiod_idx < len(returnperiods):
                        rets, tail, tail_sizes, next_returnperiod_idx, last_computed_rp, last_computed_loss = write_return_period_out(
                            next_returnperiod_idx,
                            last_computed_rp,
                            last_computed_loss,
                            retperiod,
                            value,
                            summary_id,
                            eptype,
                            sidx,
                            max_retperiod,
                            i,
                            tvar,
                            tail,
                            tail_sizes,
                            returnperiods,
                            mean_map=mean_map,
                            is_wheatsheaf=True,
                            num_sidxs=num_sidxs,
                        )
                        for ret in rets:
                            if bidx >= len(buffer):
                                yield buffer[:bidx]
                                bidx = 0
                            buffer[bidx]["SummaryId"] = ret[0]
                            buffer[bidx]["SampleId"] = ret[1]
                            buffer[bidx]["EPType"] = ret[2]
                            buffer[bidx]["ReturnPeriod"] = ret[3]
                            buffer[bidx]["Loss"] = ret[4]
                            bidx += 1
                    tvar = tvar - ((tvar - (value)) / i)
                else:
                    tvar = tvar - ((tvar - (value)) / i)

                    if idx not in tail:
                        tail[idx] = create_empty_array(TAIL_valtype)
                        tail_sizes[idx] = 0
                    tail_arr = tail[idx]
                    tail_arr = resize_array(tail_arr, tail_sizes[idx])
                    tail_current_size = tail_sizes[idx]
                    tail_arr[tail_current_size]["retperiod"] = retperiod
                    tail_arr[tail_current_size]["tvar"] = tvar
                    tail[idx] = tail_arr
                    tail_sizes[idx] += 1

                    if bidx >= len(buffer):
                        yield buffer[:bidx]
                        bidx = 0
                    buffer[bidx]["SummaryId"] = summary_id
                    buffer[bidx]["SampleId"] = sidx
                    buffer[bidx]["EPType"] = eptype
                    buffer[bidx]["ReturnPeriod"] = retperiod
                    buffer[bidx]["Loss"] = value
                    bidx += 1

                i += 1
        if use_return_period:
            unused_pw_idx = 0
            while True:
                retperiod = 0
                if unused_pw_idx < len(unused_period_weights):
                    cumulative_weighting += (
                        unused_period_weights[unused_pw_idx]["weighting"] * sample_size
                    )
                    retperiod = 1 / cumulative_weighting
                    unused_pw_idx += 1
                if next_returnperiod_idx < len(returnperiods):
                    rets, tail, tail_sizes, next_returnperiod_idx, last_computed_rp, last_computed_loss = write_return_period_out(
                        next_returnperiod_idx,
                        last_computed_rp,
                        last_computed_loss,
                        retperiod,
                        0,
                        summary_id,
                        eptype,
                        sidx,
                        max_retperiod,
                        i,
                        tvar,
                        tail,
                        tail_sizes,
                        returnperiods,
                        mean_map=mean_map,
                        is_wheatsheaf=True,
                        num_sidxs=num_sidxs,
                    )
                    for ret in rets:
                        if bidx >= len(buffer):
                            yield buffer[:bidx]
                            bidx = 0
                        buffer[bidx]["SummaryId"] = ret[0]
                        buffer[bidx]["SampleId"] = ret[1]
                        buffer[bidx]["EPType"] = ret[2]
                        buffer[bidx]["ReturnPeriod"] = ret[3]
                        buffer[bidx]["Loss"] = ret[4]
                        bidx += 1
                tvar = tvar - (tvar / i)
                i += 1
                if next_returnperiod_idx >= len(returnperiods):
                    break

    rets = write_tvar_wheatsheaf(
        num_sidxs,
        eptype_tvar,
        tail,
        tail_sizes,
    )
    for ret in rets:
        if bidx >= len(buffer):
            yield buffer[:bidx]
            bidx = 0
        buffer[bidx]["SummaryId"] = ret[0]
        buffer[bidx]["SampleId"] = ret[1]
        buffer[bidx]["EPType"] = ret[2]
        buffer[bidx]["ReturnPeriod"] = ret[3]
        buffer[bidx]["Loss"] = ret[4]
        bidx += 1
    yield buffer[:bidx]


@nb.njit(cache=True, error_model="numpy")
def write_wheatsheaf_mean(
    mean_map,
    eptype,
    epcalc,
    max_summary_id,
):
    """Generate Wheatsheaf Mean Exceedance Probability Table (EPT) by averaging losses for each return period 
    from a precomputed mean map.
    Args:
        mean_map (ndarray[MEANMAP_dtype]): An array mapping used for mean loss calculations per Summary ID.
        epcalc (int): Specifies the calculation method (mean damage loss, full uncertainty, per sample mean, sample mean).
        eptype (int): Type of exceedance probability (0 = OEP, 1 = AEP).
        max_summary_id (int): Maximum summary ID
    Yields:
        buffer (ndarray[EPT_dtype]): Buffered chunks of EPT data
    """
    if len(mean_map) == 0:
        return

    buffer = np.zeros(DEFAULT_BUFFER_SIZE, dtype=EPT_dtype)
    bidx = 0

    for summary_id in range(1, max_summary_id + 1):
        if np.sum(mean_map[summary_id - 1]["count"]) == 0:
            continue
        for mc in mean_map[summary_id - 1]:
            if bidx >= len(buffer):
                yield buffer[:bidx]
                bidx = 0
            buffer[bidx]["SummaryId"] = summary_id
            buffer[bidx]["EPCalc"] = epcalc
            buffer[bidx]["EPType"] = eptype
            buffer[bidx]["ReturnPeriod"] = mc["retperiod"]
            buffer[bidx]["Loss"] = mc["mean"] / max(mc["count"], 1)
            bidx += 1
    yield buffer[:bidx]



from oasislmf.pytools.common.data import occurrence_headers, occurrence_dtype, occurrence_fmt


headers = occurrence_headers
dtype = occurrence_dtype
fmt = occurrence_fmt
cli_support = ['bintocsv', 'csvtobin']



from oasislmf.pytools.common.data import aggregatevulnerability_headers, aggregatevulnerability_dtype, aggregatevulnerability_fmt


headers = aggregatevulnerability_headers
dtype = aggregatevulnerability_dtype
fmt = aggregatevulnerability_fmt
cli_support = ['bintocsv', 'csvtobin']



from oasislmf.pytools.common.data import vulnerability_headers, vulnerability_dtype, vulnerability_fmt


headers = vulnerability_headers
dtype = vulnerability_dtype
fmt = vulnerability_fmt
cli_support = ['bintocsv', 'csvtobin']



__all__ = [
    'get_calc_rule_ids',
    'get_grouped_fm_profile_by_level_and_term_group',
    'get_grouped_fm_terms_by_level_and_term_group',
    'get_oed_hierarchy',
    'get_il_input_items',
]

import contextlib
import copy
import itertools
import os
import time
import warnings

import numpy as np
import pandas as pd
from pandas.api.types import is_numeric_dtype
from ods_tools.oed import fill_empty, BLANK_VALUES

from oasislmf.preparation.summaries import get_useful_summary_cols
from oasislmf.pytools.common.data import (fm_policytc_headers, fm_policytc_dtype,
                                          fm_profile_headers, fm_profile_dtype,
                                          fm_profile_step_headers, fm_profile_step_dtype,
                                          fm_programme_headers, fm_programme_dtype,
                                          fm_xref_headers, fm_xref_dtype
                                          )
from oasislmf.utils.calc_rules import get_calc_rules
from oasislmf.utils.coverages import SUPPORTED_COVERAGE_TYPES
from oasislmf.utils.data import factorize_ndarray
from oasislmf.utils.defaults import (OASIS_FILES_PREFIXES,
                                     get_default_accounts_profile, get_default_exposure_profile,
                                     get_default_fm_aggregation_profile, SOURCE_IDX)
from oasislmf.utils.exceptions import OasisException
from oasislmf.utils.fm import (CALCRULE_ASSIGNMENT_METHODS, COVERAGE_AGGREGATION_METHODS,
                               DEDUCTIBLE_AND_LIMIT_TYPES, FM_LEVELS, FML_ACCALL, STEP_TRIGGER_TYPES,
                               SUPPORTED_FM_LEVELS, FM_TERMS, GROUPED_SUPPORTED_FM_LEVELS)
from oasislmf.utils.log import oasis_log
from oasislmf.utils.path import as_path
from oasislmf.utils.profiles import (get_default_step_policies_profile,
                                     get_grouped_fm_profile_by_level_and_term_group,
                                     get_grouped_fm_terms_by_level_and_term_group,
                                     get_oed_hierarchy)

pd.options.mode.chained_assignment = None
warnings.simplefilter(action='ignore', category=FutureWarning)

# convertion from np dtype to pandas dtype
fm_policytc_dtype = {col: dtype for col, (dtype, _) in fm_policytc_dtype.fields.items()}
fm_profile_dtype = {col: dtype for col, (dtype, _) in fm_profile_dtype.fields.items()}
fm_profile_step_dtype = {col: dtype for col, (dtype, _) in fm_profile_step_dtype.fields.items()}
fm_programme_dtype = {col: dtype for col, (dtype, _) in fm_programme_dtype.fields.items()}
fm_xref_dtype = {col: dtype for col, (dtype, _) in fm_xref_dtype.fields.items()}


# Define a list of all supported OED coverage types in the exposure
supp_cov_type_ids = [v['id'] for v in SUPPORTED_COVERAGE_TYPES.values()]

policytc_cols = ['profile_id', 'calcrule_id', 'deductible', 'deductible_min', 'deductible_max', 'attachment', 'limit', 'share']

profile_cols_map = {
    'deductible': 'deductible1',
    'deductible_min': 'deductible2',
    'deductible_max': 'deductible3',
    'attachment': 'attachment1',
    'limit': 'limit1',
    'share': 'share1'
}
cross_layer_level = {FML_ACCALL, }

risk_disaggregation_term = {'deductible', 'deductible_min', 'deductible_max', 'attachment', 'limit'}

fm_term_ids = [fm_term['id'] for fm_term in FM_TERMS.values()]


def prepare_ded_and_limit(level_df):
    simplify_no_terms = {
        'deductible': ['ded_type', 'ded_code'],
        'limit': ['lim_type', 'lim_code']
    }
    level_df['need_tiv'] = False
    for term, type_and_code in simplify_no_terms.items():
        if term in level_df.columns:
            for elm in type_and_code:
                if elm in level_df.columns:
                    level_df.loc[level_df[term] == 0, elm] = 0
                    if elm[-5:] == '_type':
                        level_df['need_tiv'] |= level_df[elm].isin({2, 3})  # 2 or 3 type means term depend on tiv
        else:
            level_df = level_df.drop(columns=type_and_code, errors='ignore')

    return level_df


def get_calc_rule_ids(il_inputs_calc_rules_df, calc_rule_type):
    """
    merge selected il_inputs with the correct calc_rule table and  return a pandas Series of calc. rule IDs

    Args:
     il_inputs_calc_rules_df (DataFrame):  IL input items dataframe
     calc_rule_type (str): type of calc_rule to look for

    Returns:
        pandas Series of calc. rule IDs
    """
    il_inputs_calc_rules_df = il_inputs_calc_rules_df.copy()
    calc_rules_df, calc_rule_term_info = get_calc_rules(calc_rule_type)
    calc_rules_df = calc_rules_df.drop(columns=['desc', 'id_key'], axis=1, errors='ignore')

    terms = []
    terms_indicators = []
    no_terms = {
        'deductible': ['ded_type', 'ded_code'],
        'limit': ['lim_type', 'lim_code']
    }
    for term in calc_rule_term_info['terms']:
        if term in il_inputs_calc_rules_df.columns:
            terms.append(term)
            terms_indicators.append('{}_gt_0'.format(term))
        else:
            calc_rules_df = calc_rules_df[calc_rules_df['{}_gt_0'.format(term)] == 0].drop(columns=['{}_gt_0'.format(term)])

    for term in calc_rule_term_info['types_and_codes']:
        if term in il_inputs_calc_rules_df.columns:
            il_inputs_calc_rules_df[term] = il_inputs_calc_rules_df[term].fillna(0).astype('uint8')
        else:
            calc_rules_df = calc_rules_df[calc_rules_df[term] == 0].drop(columns=[term])

    for term, type_and_code in no_terms.items():
        if f'{term}_gt_0' in calc_rules_df.columns:
            for elm in type_and_code:
                if elm in il_inputs_calc_rules_df.columns:
                    il_inputs_calc_rules_df.loc[il_inputs_calc_rules_df[term] == 0, elm] = 0
        else:
            for elm in type_and_code:
                if elm in calc_rules_df.columns:
                    calc_rules_df = calc_rules_df[calc_rules_df[elm] == 0].drop(columns=[elm])

    il_inputs_calc_rules_df.loc[:, terms_indicators] = np.where(il_inputs_calc_rules_df[terms] > 0, 1, 0)

    merge_col = list(set(il_inputs_calc_rules_df.columns).intersection(calc_rules_df.columns).difference({'calcrule_id'}))
    if len(merge_col):
        calcrule_ids = (
            il_inputs_calc_rules_df[merge_col].reset_index()
            .merge(calc_rules_df[merge_col + ['calcrule_id']].drop_duplicates(), how='left', on=merge_col)
        ).set_index('index')['calcrule_id'].fillna(0)
    else:
        return 100  # no term we return pass through
    if 0 in calcrule_ids.unique():
        _cols = list(set(['PortNumber', 'AccNumber', 'LocNumber'] + merge_col).intersection(il_inputs_calc_rules_df.columns))
        no_match_keys = il_inputs_calc_rules_df.loc[calcrule_ids == 0, _cols].drop_duplicates()
        err_msg = 'Calculation Rule mapping error, non-matching keys:\n{}'.format(no_match_keys)
        raise OasisException(err_msg)
    return calcrule_ids


def get_profile_ids(il_inputs_df):
    """
    Returns a Numpy array of policy TC IDs from a table of IL input items

    :param il_inputs_df: IL input items dataframe
    :type il_inputs_df: pandas.DataFrame

    :return: Numpy array of policy TC IDs
    :rtype: numpy.ndarray
    """
    factor_col = list(set(il_inputs_df.columns).intersection(policytc_cols).difference({'profile_id', }))
    return factorize_ndarray(il_inputs_df.loc[:, factor_col].values, col_idxs=range(len(factor_col)))[0]


def __split_fm_terms_by_risk(df):
    """
    adjust the financial term to the number of risks
    for example deductible is split into each individul building risk

    Args:
         df (DataFrame): the DataFrame an FM level
    """
    for term in risk_disaggregation_term.intersection(set(df.columns)):
        if f'{term[:3]}_code' in df.columns:
            code_filter = df[f'{term[:3]}_code'] == 0
            df.loc[code_filter, term] /= df.loc[code_filter, 'NumberOfRisks']
        else:
            df[term] /= df['NumberOfRisks']


def get_cond_info(locations_df, accounts_df):
    pol_info = {}
    level_conds = {}
    extra_accounts = []
    default_cond_tag = '0'
    if 'CondTag' in locations_df.columns:
        fill_empty(locations_df, 'CondTag', default_cond_tag)
        loc_condkey_df = locations_df.loc[locations_df['CondTag'] != default_cond_tag, ['PortNumber', 'AccNumber', 'CondTag']].drop_duplicates()
    else:
        loc_condkey_df = pd.DataFrame([], columns=['PortNumber', 'AccNumber', 'CondTag'])

    if 'CondTag' in accounts_df.columns:
        fill_empty(accounts_df, 'CondTag', default_cond_tag)
        acc_condkey_df = accounts_df.loc[accounts_df['CondTag'] != '', ['PortNumber', 'AccNumber', 'CondTag']].drop_duplicates()
        condkey_match_df = acc_condkey_df.merge(loc_condkey_df, how='outer', indicator=True)
        missing_condkey_df = condkey_match_df.loc[condkey_match_df['_merge'] == 'right_only', ['PortNumber', 'AccNumber', 'CondTag']]
    else:
        acc_condkey_df = pd.DataFrame([], columns=['PortNumber', 'AccNumber', 'CondTag'])
        missing_condkey_df = loc_condkey_df

    if missing_condkey_df.shape[0]:
        raise OasisException(f'Those condtag are present in locations but missing in the account file:\n{missing_condkey_df}')

    if acc_condkey_df.shape[0]:
        if 'CondTag' not in locations_df.columns:
            locations_df['CondTag'] = default_cond_tag
        # we get information about cond from accounts_df
        cond_tags = {}  # information about each cond tag
        account_layer_exclusion = {}  # for each account and layer, store info about cond class exclusion
        if 'CondPriority' in accounts_df.columns:
            fill_empty(accounts_df, 'CondPriority', 1)
        else:
            accounts_df['CondPriority'] = 1
        if 'CondPeril' in accounts_df.columns:
            fill_empty(accounts_df, 'CondPeril', '')
        else:
            accounts_df['CondPeril'] = ''
        for acc_rec in accounts_df.to_dict(orient="records"):
            cond_tag_key = (acc_rec['PortNumber'], acc_rec['AccNumber'], acc_rec['CondTag'])
            cond_number_key = (acc_rec['PortNumber'], acc_rec['AccNumber'], acc_rec['CondTag'], acc_rec['CondNumber'])
            cond_tag = cond_tags.setdefault(cond_tag_key, {'CondPriority': acc_rec['CondPriority'] or 1, 'CondPeril': acc_rec['CondPeril']})
            cond_tag.setdefault('layers', {})[acc_rec['layer_id']] = {'CondNumber': cond_number_key}
            exclusion_cond_tags = account_layer_exclusion.setdefault((acc_rec['PortNumber'], acc_rec['AccNumber']), {}).setdefault(acc_rec['layer_id'],
                                                                                                                                   set())
            pol_info[(acc_rec['PortNumber'], acc_rec['AccNumber'], acc_rec['layer_id'])] = [acc_rec['PolNumber'], acc_rec['LayerNumber']]
            if acc_rec.get('CondClass') == 1:
                exclusion_cond_tags.add(acc_rec['CondTag'])

        # we get the list of loc for each cond_tag
        loc_conds = {}
        KEY_INDEX = 0
        PRIORITY_INDEX = 1
        for loc_rec in locations_df.to_dict(orient="records"):
            loc_key = (loc_rec['PortNumber'], loc_rec['AccNumber'], loc_rec['LocNumber'])
            cond_key = (loc_rec['PortNumber'], loc_rec['AccNumber'], loc_rec.get('CondTag', default_cond_tag))
            if cond_key in cond_tags:
                cond_tag = cond_tags[cond_key]
            else:
                cond_tag = {'CondPriority': 1, 'layers': {}}
                cond_tags[cond_key] = cond_tag

            cond_location = cond_tag.setdefault('locations', set())
            cond_location.add(loc_key)
            cond_tag_priority = cond_tag['CondPriority']
            conds = loc_conds.setdefault(loc_key, [])

            for i, cond in enumerate(conds):
                if cond_tag_priority < cond[PRIORITY_INDEX]:
                    conds.insert(i, (cond_key, cond_tag_priority))
                    break
                elif cond_tag_priority == cond[PRIORITY_INDEX] and cond_key != cond[KEY_INDEX]:
                    raise OasisException(f"{cond_key} and {cond[KEY_INDEX]} have same priority in {loc_key}")
            else:
                conds.append((cond_key, cond_tag_priority))

        # at first we just want condtag for each level
        for cond_key, cond_info in cond_tags.items():
            port_num, acc_num, cond_tag = cond_key
            cond_level_start = 1
            for loc_key in cond_info.get('locations', set()):
                for i, (cond_key_i, _) in enumerate(loc_conds[loc_key]):
                    if cond_key_i == cond_key:
                        cond_level_start = max(cond_level_start, i + 1)
                        break
            cond_info['cond_level_start'] = cond_level_start
            for layer_id, exclusion_conds in account_layer_exclusion[(cond_key[0], cond_key[1])].items():
                if layer_id not in cond_info['layers']:
                    PolNumber, LayerNumber = pol_info[(port_num, acc_num, layer_id)]
                    if exclusion_conds:
                        extra_accounts.append({
                            'PortNumber': port_num,
                            'AccNumber': acc_num,
                            'PolNumber': PolNumber,
                            'LayerNumber': LayerNumber,
                            'CondTag': cond_tag,
                            'layer_id': layer_id,
                            'CondNumber': 'FullFilter',
                            'CondDed6All': 1,
                            'CondDedType6All': 1,
                            'CondPeril': 'AA1',
                        })
                    else:
                        extra_accounts.append({
                            'PortNumber': port_num,
                            'AccNumber': acc_num,
                            'PolNumber': PolNumber,
                            'LayerNumber': LayerNumber,
                            'CondTag': cond_tag,
                            'layer_id': layer_id,
                            'CondNumber': '',
                            'CondPeril': 'AA1',
                        })
            level_conds.setdefault(cond_level_start, set()).add(cond_key)
    return level_conds, extra_accounts


def get_levels(locations_df, accounts_df):
    if locations_df is not None and not {"CondTag", "CondNumber"}.difference(accounts_df.columns):
        level_conds, extra_accounts = get_cond_info(locations_df, accounts_df)
    else:
        level_conds = []
    for group_name, group_info in copy.deepcopy(GROUPED_SUPPORTED_FM_LEVELS).items():
        if group_info['oed_source'] == 'location':
            if locations_df is not None:
                locations_df['layer_id'] = 1
                yield locations_df, list(group_info['levels'].items()), group_info.get('fm_peril_field'), None
        elif group_info['oed_source'] == 'account':
            if group_name == 'cond' and level_conds:
                loc_conds_df = locations_df[['loc_id', 'PortNumber', 'AccNumber', 'CondTag']].drop_duplicates()
                for stage, cond_keys in level_conds.items():
                    cond_filter_df = (pd.DataFrame(cond_keys, columns=['PortNumber', 'AccNumber', 'CondTag'])
                                      .sort_values(by=['PortNumber', 'AccNumber', 'CondTag']))
                    loc_conds_df_filter = cond_filter_df[['PortNumber', 'AccNumber', 'CondTag']].drop_duplicates().merge(loc_conds_df, how='left')
                    yield (cond_filter_df.merge(pd.concat([accounts_df, pd.DataFrame(extra_accounts)]), how='left'),
                           group_info['levels'].items(),
                           group_info['fm_peril_field'],
                           loc_conds_df_filter[['loc_id', 'CondTag']])
            else:
                yield accounts_df, group_info['levels'].items(), group_info.get('fm_peril_field'), None


def get_level_term_info(term_df_source, level_column_mapper, level_id, step_level, fm_peril_field, oed_schema):
    level_terms = set()
    terms_maps = {}
    coverage_group_map = {}
    fm_group_tiv = {}
    non_zero_default = {}
    for ProfileElementName, term_info in level_column_mapper[level_id].items():
        if term_info.get("FMTermType") == "tiv":
            continue
        default_value = oed_schema.get_default(ProfileElementName)
        if ProfileElementName not in term_df_source.columns:
            if default_value == 0 or default_value in BLANK_VALUES:
                continue
            else:
                non_zero_default[ProfileElementName] = [term_info['FMTermType'].lower(), default_value]
                continue
        else:
            fill_empty(term_df_source, ProfileElementName, default_value)

        if pd.isna(default_value):
            non_default_val = ~term_df_source[ProfileElementName].isna()
        else:
            non_default_val = (term_df_source[ProfileElementName] != default_value)
        if 'FMProfileStep' in term_info:
            profile_steps = term_info["FMProfileStep"]
            if isinstance(profile_steps, int):
                profile_steps = [profile_steps]
            valid_step_trigger_types = term_df_source.loc[(term_df_source['StepTriggerType'].isin(profile_steps))
                                                          & non_default_val, 'StepTriggerType'].unique()
            if len(valid_step_trigger_types):
                level_terms.add(term_info['FMTermType'].lower())
            for step_trigger_type in valid_step_trigger_types:
                coverage_aggregation_method = STEP_TRIGGER_TYPES[step_trigger_type]['coverage_aggregation_method']
                calcrule_assignment_method = STEP_TRIGGER_TYPES[step_trigger_type]['calcrule_assignment_method']
                for coverage_type_id in supp_cov_type_ids:
                    FMTermGroupID = COVERAGE_AGGREGATION_METHODS[coverage_aggregation_method].get(coverage_type_id)
                    if FMTermGroupID is None:  # step policy not supported for this coverage
                        continue

                    if (step_trigger_type, coverage_type_id) in coverage_group_map:
                        if coverage_group_map[(step_trigger_type, coverage_type_id)] != FMTermGroupID:
                            raise OasisException(
                                f"multiple coverage_type_id {(step_trigger_type, coverage_type_id)} "
                                f"assigned to the different FMTermGroupID "
                                f"{(FMTermGroupID, coverage_group_map[(step_trigger_type, coverage_type_id)])}")
                    else:
                        coverage_group_map[(step_trigger_type, coverage_type_id)] = FMTermGroupID

                    terms_map = terms_maps.setdefault((FMTermGroupID, step_trigger_type), {fm_peril_field: 'fm_peril'} if fm_peril_field else {})
                    if CALCRULE_ASSIGNMENT_METHODS[calcrule_assignment_method][FMTermGroupID]:
                        terms_map[ProfileElementName] = term_info['FMTermType'].lower()
        else:
            if not (non_default_val).any():
                continue
            level_terms.add(term_info['FMTermType'].lower())
            coverage_type_ids = term_info.get("CoverageTypeID", supp_cov_type_ids)
            FMTermGroupID = term_info.get('FMTermGroupID', 1)
            if step_level:
                term_key = (FMTermGroupID, 0)
            else:
                term_key = FMTermGroupID

            if isinstance(coverage_type_ids, int):
                coverage_type_ids = [coverage_type_ids]
            fm_group_tiv[FMTermGroupID] = coverage_type_ids
            for coverage_type_id in coverage_type_ids:
                if step_level:
                    coverage_group_key = (0, coverage_type_id)
                else:
                    coverage_group_key = coverage_type_id
                if coverage_type_id in coverage_group_map:
                    if coverage_group_map[coverage_group_key] != FMTermGroupID:
                        raise OasisException(
                            f"multiple coverage_type_id {coverage_group_key}"
                            f"assigned to the different FMTermGroupID {(FMTermGroupID, coverage_group_map[coverage_group_key])}")
                else:
                    coverage_group_map[coverage_group_key] = FMTermGroupID
            terms_maps.setdefault(term_key, {fm_peril_field: 'fm_peril'} if fm_peril_field else {})[
                ProfileElementName] = term_info['FMTermType'].lower()
    if level_terms:
        for ProfileElementName, (term, default_value) in non_zero_default.items():
            term_df_source[ProfileElementName] = default_value
            level_terms.add(term)
            for terms_map in terms_maps.values():
                terms_map[ProfileElementName] = term

    return level_terms, terms_maps, coverage_group_map, fm_group_tiv


def associate_items_peril_to_policy_peril(item_perils, policy_df, fm_peril_col, oed_schema):
    """
    for each peril_id in item_peril_list we map it to each string representing policy perils
    we then merge this mapping to the initial policies so that each line will have a peril_id
    that can be use directly as key when merging with the gul_input_df
    Args:
        item_perils: all peril id from gul_input_df
        policy_df: the df with the policies
        fm_peril_col: the name of the column to use that define the policy peril filter
        oed_schema: the schema object we use to map peril and subperils
    Returns:

    """
    fm_perils = policy_df[[fm_peril_col]].drop_duplicates()
    peril_map_df = pd.merge(fm_perils, item_perils, how='cross')
    peril_map_df = peril_map_df[oed_schema.peril_filtering(peril_map_df['peril_id'], peril_map_df[fm_peril_col])]
    return policy_df.merge(peril_map_df)


@oasis_log
def get_il_input_items(
        gul_inputs_df,
        exposure_data,
        target_dir,
        logger,
        exposure_profile=get_default_exposure_profile(),
        accounts_profile=get_default_accounts_profile(),
        fm_aggregation_profile=get_default_fm_aggregation_profile(),
        do_disaggregation=True,
        oasis_files_prefixes=OASIS_FILES_PREFIXES['il'],
        chunksize=(2 * 10 ** 5),
):
    """
    Generates and returns a Pandas dataframe of IL input items.

    :param gul_inputs_df: GUL input items
    :type gul_inputs_df: pandas.DataFrame

    :param exposure_data: object containing all information about the insurance policies

    :param target_dir: path to the directory used to write the fm files

    :param logger: logger object to trace progress

    :param exposure_profile: Source exposure profile (optional)
    :type exposure_profile: dict

    :param accounts_profile: Source accounts profile (optional)
    :type accounts_profile: dict

    :param fm_aggregation_profile: FM aggregation profile (optional)
    :param fm_aggregation_profile: dict

    :param do_disaggregation: whether to split terms and conditions for aggregate exposure (optional)
    :param do_disaggregation: bool

    :return: IL inputs dataframe
    :rtype: pandas.DataFrame

    :return Accounts dataframe
    :rtype: pandas.DataFrame
    """
    target_dir = as_path(target_dir, 'Target IL input files directory', is_dir=True, preexists=False)
    with contextlib.ExitStack() as stack:
        fm_policytc_file = stack.enter_context(open(os.path.join(target_dir, f"{oasis_files_prefixes['fm_policytc']}.csv"), 'w'))
        fm_policytc_file.write(f"{','.join(fm_policytc_headers)}{os.linesep}")

        fm_profile_file = stack.enter_context(open(os.path.join(target_dir, f"{oasis_files_prefixes['fm_profile']}.csv"), 'w'))

        fm_programme_file = stack.enter_context(open(os.path.join(target_dir, f"{oasis_files_prefixes['fm_programme']}.csv"), 'w'))
        fm_programme_file.write(f"{','.join(fm_programme_headers)}{os.linesep}")

        fm_xref_file = stack.enter_context(open(os.path.join(target_dir, f"{oasis_files_prefixes['fm_xref']}.csv"), 'w'))

        locations_df = exposure_data.location.dataframe if exposure_data.location is not None else None
        accounts_df = exposure_data.account.dataframe
        oed_schema = exposure_data.oed_schema

        profile = get_grouped_fm_profile_by_level_and_term_group(exposure_profile, accounts_profile)

        # Get the FM aggregation profile - this describes how the IL input
        # items are to be aggregated in the various FM levels
        fm_aggregation_profile = copy.deepcopy(fm_aggregation_profile)

        if not fm_aggregation_profile:
            raise OasisException(
                'FM aggregation profile is empty - this is required to perform aggregation'
            )
        # Get the OED hierarchy terms profile - this defines the column names for loc.
        # ID, acc. ID, policy no. and portfolio no., as used in the source exposure
        # and accounts files. This is to ensure that the method never makes hard
        # coded references to the corresponding columns in the source files, as
        # that would mean that changes to these column names in the source files
        # may break the method
        oed_hierarchy = get_oed_hierarchy(exposure_profile, accounts_profile)
        tiv_terms = {v['tiv']['ProfileElementName']: str(v['tiv']['CoverageTypeID']) for k, v in
                     profile[FM_LEVELS['site coverage']['id']].items()}
        useful_cols = sorted(set(['layer_id', 'orig_level_id', 'level_id', 'agg_id', 'gul_input_id', 'tiv', 'NumberOfRisks']
                                 + get_useful_summary_cols(oed_hierarchy)).union(tiv_terms)
                             - {'profile_id', 'item_id', 'output_id'}, key=str.lower)
        gul_inputs_df = gul_inputs_df.rename(columns={'item_id': 'gul_input_id'})
        # adjust tiv columns and name them as their coverage id
        gul_inputs_df[['risk_id', 'NumberOfRisks']] = gul_inputs_df[['building_id', 'NumberOfBuildings']]
        gul_inputs_df.loc[gul_inputs_df['IsAggregate'] == 0, 'risk_id'] = 1
        gul_inputs_df.loc[gul_inputs_df['IsAggregate'] == 0, 'NumberOfRisks'] = 1
        gul_inputs_df.loc[gul_inputs_df['NumberOfRisks'] == 0, 'NumberOfRisks'] = 1

        # initialization
        agg_keys = set()
        for level_id in fm_aggregation_profile:
            agg_keys = agg_keys.union(set([v['field'] for v in fm_aggregation_profile[level_id]['FMAggKey'].values()]))

        present_cols = [col for col in gul_inputs_df.columns if col in set(useful_cols).union(agg_keys).union(fm_term_ids)]
        gul_inputs_df = gul_inputs_df[present_cols].rename(columns=tiv_terms)

        # Profile dict are base on key that correspond to the fm term name.
        # this prevents multiple file columns to point to the same fm term
        # which is necessary to have a generic logic that works with step policy
        # so we change the key to be the column and use FMTermType to store the term name
        level_column_mapper = {}
        for level_id, level_profile in profile.items():
            column_map = {}
            level_column_mapper[level_id] = column_map

            # for fm we only use term_id 1
            for term_name, term_info in itertools.chain.from_iterable(profile.items() for profile in level_profile.values()):
                new_term_info = copy.deepcopy(term_info)
                new_term_info['FMTermType'] = term_name
                column_map[term_info['ProfileElementName']] = new_term_info

        gul_inputs_df = gul_inputs_df.drop(columns=fm_term_ids, errors='ignore').reset_index(drop=True)
        gul_inputs_df['agg_id_prev'] = gul_inputs_df['gul_input_id']
        gul_inputs_df['layer_id'] = 1
        gul_inputs_df['PolNumber'] = pd.NA
        gul_inputs_df = gul_inputs_df.rename(columns={})
        item_perils = gul_inputs_df[['peril_id']].drop_duplicates()

        # Determine whether step policies are listed, are not full of nans and step
        # numbers are greater than zero
        step_policies_present = False

        extra_fm_col = ['layer_id', 'PolNumber', 'NumberOfBuildings', 'IsAggregate']
        if 'StepTriggerType' in accounts_df and 'StepNumber' in accounts_df:
            fill_empty(accounts_df, ['StepTriggerType', 'StepNumber'], 0)
            step_account = (accounts_df[accounts_df[accounts_df['StepTriggerType'].notnull()]['StepNumber'].gt(0)][['PortNumber', 'AccNumber']]
                            .drop_duplicates())
            step_policies_present = bool(step_account.shape[0])
            # this is done only for fmcalc to make sure it program nodes stay as a tree, remove 'is_step' logic when fmcalc is dropped
            if step_policies_present:
                step_account['is_step'] = 1
                gul_inputs_df = gul_inputs_df.merge(step_account, how='left')
                gul_inputs_df['is_step'] = gul_inputs_df['is_step'].fillna(0).astype('int8')
                extra_fm_col.append('is_step')
        # If step policies listed, keep step trigger type and columns associated
        # with those step trigger types that are present
        if step_policies_present:
            # we happend the fm step policy term to policy layer
            step_policy_level_map = level_column_mapper[SUPPORTED_FM_LEVELS['policy all']['id']]
            for col in ['StepTriggerType', 'cov_agg_id', 'assign_step_calcrule']:
                step_policy_level_map[col] = {
                    'ProfileElementName': col,
                    'FMTermType': col,
                }
            for key, step_term in get_default_step_policies_profile().items():
                step_policy_level_map[step_term['Key']] = {
                    'ProfileElementName': step_term['Key'],
                    'FMTermType': step_term['FMProfileField'],
                    'FMProfileStep': step_term.get('FMProfileStep')
                }
            fm_profile_cols = fm_profile_step_headers
        else:
            fm_profile_cols = fm_profile_headers

        fm_profile_file.write(f"{','.join(fm_profile_cols)}{os.linesep}")
        pass_through_profile = pd.DataFrame({col: [0] for col in fm_profile_cols})
        pass_through_profile['profile_id'] = 1
        pass_through_profile['calcrule_id'] = 100
        write_fm_profile_level(pass_through_profile, fm_profile_file, step_policies_present, chunksize)

        profile_id_offset = 1  # profile_id 1 is the passthrough policy 100
        cur_level_id = 0
        for term_df_source, levels, fm_peril_field, CondTag_merger in get_levels(locations_df, accounts_df):
            t0 = time.time()
            if CondTag_merger is not None:
                gul_inputs_df = gul_inputs_df.drop(columns='CondTag', errors='ignore').merge(CondTag_merger, how='left')

            for level, level_info in levels:
                level_id = level_info['id']
                is_policy_layer_level = level_id == SUPPORTED_FM_LEVELS['policy layer']['id']
                step_level = 'StepTriggerType' in level_column_mapper[level_id]  # only true is step policy are present
                level_terms, terms_maps, coverage_group_map, fm_group_tiv = get_level_term_info(
                    term_df_source, level_column_mapper, level_id, step_level, fm_peril_field, oed_schema)
                agg_key = [v['field'] for v in fm_aggregation_profile[level_id]['FMAggKey'].values()]
                if not terms_maps:  # no terms we skip this level
                    if is_policy_layer_level:  # for policy layer we group all to make sure we can have a0 ALLOCATION_RULE
                        cur_level_id += 1
                        write_empty_policy_layer(gul_inputs_df, cur_level_id, agg_key, fm_policytc_file,
                                                 fm_programme_file, chunksize)
                        logger.info(f"level {cur_level_id} {level_info} took {time.time() - t0}")
                        t0 = time.time()
                    continue

                # reset non layered agg_id
                layered_agg_id = gul_inputs_df[gul_inputs_df['layer_id'] > 1]["agg_id_prev"].unique()
                gul_inputs_df['layer_id'] = gul_inputs_df['layer_id'].where(gul_inputs_df['agg_id_prev'].isin(layered_agg_id), 0)

                # get all rows with terms in term_df_source and determine the correct FMTermGroupID
                level_df_list = []
                valid_term_default = {}
                for group_key, terms in terms_maps.items():
                    if step_level:
                        FMTermGroupID, step_trigger_type = group_key
                        group_df = term_df_source[term_df_source['StepTriggerType'] == step_trigger_type]
                        terms = {**terms_maps.get((1, 0), {}), **terms}  # take all common terms in (1,0) plus term with step_trigger_type filter
                    else:
                        FMTermGroupID = group_key
                        group_df = term_df_source
                    group_df = (group_df[list(set(agg_key + list(terms.keys()) + extra_fm_col)
                                              .union(set(useful_cols).difference(set(gul_inputs_df.columns)))
                                              .intersection(group_df.columns))]
                                .assign(FMTermGroupID=FMTermGroupID))

                    # only keep policy with non default values, remove duplicate
                    numeric_terms = [term for term in terms.keys() if is_numeric_dtype(group_df[term])]
                    term_filter = False
                    for term in numeric_terms:
                        if pd.isna(oed_schema.get_default(term)):
                            term_filter |= ~group_df[term].isna()
                            valid_term_default[terms[term]] = 0
                        else:
                            term_filter |= (group_df[term] != oed_schema.get_default(term))
                            valid_term_default[terms[term]] = oed_schema.get_default(term)
                        term_filter |= (group_df[term] != oed_schema.get_default(term))
                    keep_df = group_df[term_filter][list(
                        set(agg_key).intersection(group_df.columns))].drop_duplicates()

                    group_df = group_df.merge(keep_df, how='inner')

                    # multiple ProfileElementName can have the same fm terms (ex: StepTriggerType 5), we take the max to have a unique one
                    for ProfileElementName, term in terms.items():
                        if term in group_df:
                            group_df[term] = group_df[[term, ProfileElementName]].max(axis=1)
                            group_df.drop(columns=ProfileElementName, inplace=True)
                        else:
                            group_df.rename(columns={ProfileElementName: term}, inplace=True)
                    level_df_list.append(group_df)
                level_df = pd.concat(level_df_list, copy=True, ignore_index=True)
                base_acc = level_df['AccNumber']

                for term, default in valid_term_default.items():
                    level_df[term] = level_df[term].fillna(default)

                if do_disaggregation and 'risk_id' in agg_key:
                    level_df['NumberOfRisks'] = level_df['NumberOfBuildings'].mask(level_df['IsAggregate'] == 0, 1)
                    __split_fm_terms_by_risk(level_df)
                    level_df = level_df.drop(columns=['NumberOfBuildings', 'IsAggregate', 'NumberOfRisks'])

                level_df = level_df.drop_duplicates(subset=set(level_df.columns) - set(SOURCE_IDX.values()))
                agg_id_merge_col = agg_key + ['FMTermGroupID']
                agg_id_merge_col_extra = ['need_tiv']
                if step_level:
                    # merge with gul_inputs_df needs to be based on 'steptriggertype' and 'coverage_type_id'
                    coverage_type_id_df = pd.DataFrame(
                        [[StepTriggerType, coverage_type_id, FMTermGroupID]
                            for (StepTriggerType, coverage_type_id), FMTermGroupID in coverage_group_map.items()],
                        columns=['steptriggertype', 'coverage_type_id', 'FMTermGroupID'])
                    level_df = level_df.merge(coverage_type_id_df)
                    level_df["is_step"] = level_df['steptriggertype'] > 0
                    agg_id_merge_col_extra.append('coverage_type_id')
                else:
                    # map the coverage_type_id to the correct FMTermGroupID for this level. coverage_type_id without term and therefor FMTermGroupID are map to 0
                    gul_inputs_df['FMTermGroupID'] = (gul_inputs_df['coverage_type_id']
                                                      .map(coverage_group_map, na_action='ignore')
                                                      .astype('Int32'))

                # check that peril_id is part of the fm peril policy, if not we remove the terms
                if 'fm_peril' in level_df:
                    level_df = associate_items_peril_to_policy_peril(item_perils, level_df, 'fm_peril', oed_schema)
                    factorize_key = agg_key + ['FMTermGroupID', 'fm_peril']
                    agg_id_merge_col += ['fm_peril']
                    agg_id_merge_col_extra.append('peril_id')
                else:
                    factorize_key = agg_key + ['FMTermGroupID']

                if level_df.empty:  # No actual terms for this level
                    if is_policy_layer_level:  # for policy layer we group all to make sure we can have a0 ALLOCATION_RULE
                        cur_level_id += 1
                        write_empty_policy_layer(gul_inputs_df, cur_level_id, agg_key, fm_policytc_file,
                                                 fm_programme_file, chunksize)
                        logger.info(f"level {cur_level_id} {level_info} took {time.time() - t0}")
                        t0 = time.time()
                    continue

                level_df = prepare_ded_and_limit(level_df)

                agg_id_merge_col = list(set(agg_id_merge_col).intersection(level_df.columns))
                gul_inputs_df = gul_inputs_df.merge(
                    level_df[agg_id_merge_col + agg_id_merge_col_extra].drop_duplicates(), how='left', validate='many_to_one')
                if is_policy_layer_level:  # we merge all on account at this level even if there is no policy
                    gul_inputs_df["FMTermGroupID"] = gul_inputs_df["FMTermGroupID"].fillna(-1).astype('i4')
                else:
                    gul_inputs_df["FMTermGroupID"] = gul_inputs_df["FMTermGroupID"].mask(
                        gul_inputs_df["need_tiv"].isna(), -gul_inputs_df["agg_id_prev"]).astype('i4')
                gul_inputs_df["agg_id"] = factorize_ndarray(gul_inputs_df.loc[:, factorize_key].values, col_idxs=range(len(factorize_key)))[0]

                # merge Tiv in level_df when needed
                # make sure correct tiv sum exist
                tiv_df_list = []
                for FMTermGroupID, coverage_type_ids in fm_group_tiv.items():
                    tiv_key = '_'.join(map(str, sorted(coverage_type_ids)))
                    if tiv_key not in gul_inputs_df:
                        gul_inputs_df[tiv_key] = gul_inputs_df[list(
                            set(gul_inputs_df.columns).intersection(map(str, sorted(coverage_type_ids))))].sum(axis=1)

                    tiv_df_list.append(gul_inputs_df[(gul_inputs_df["need_tiv"] == True) &
                                                     (gul_inputs_df["FMTermGroupID"] == FMTermGroupID)]
                                       .drop_duplicates(subset=['agg_id', 'loc_id', 'building_id'])
                                       .rename(columns={tiv_key: 'agg_tiv'})
                                       .groupby("agg_id", observed=True)
                                       .agg({**{col: 'first' for col in agg_id_merge_col}, **{'agg_tiv': 'sum'}})
                                       )
                tiv_df = pd.concat(tiv_df_list)

                gul_inputs_df = gul_inputs_df.merge(tiv_df['agg_tiv'], left_on='agg_id', right_index=True, how='left')
                gul_inputs_df['agg_tiv'] = gul_inputs_df['agg_tiv'].fillna(0)

                level_df = level_df.merge(tiv_df.drop_duplicates(), how='left').drop(columns=['need_tiv'])
                level_df['agg_tiv'] = level_df['agg_tiv'].fillna(0)
                # Apply rule to convert type 2 deductibles and limits to TIV shares
                if 'deductible' in level_df.columns and 'ded_type' in level_df.columns:
                    level_df['deductible'] = level_df['deductible'].fillna(0.)
                    level_df['ded_type'] = level_df['ded_type'].infer_objects(copy=False).fillna(0.).astype('i4')
                    level_df['deductible'] = np.where(
                        level_df['ded_type'] == DEDUCTIBLE_AND_LIMIT_TYPES['pctiv']['id'],
                        level_df['deductible'] * level_df['agg_tiv'],
                        level_df['deductible']
                    )
                    level_df.loc[level_df['ded_type'] == DEDUCTIBLE_AND_LIMIT_TYPES['pctiv']
                                 ['id'], 'ded_type'] = DEDUCTIBLE_AND_LIMIT_TYPES['flat']['id']

                    type_filter = level_df['ded_type'] == DEDUCTIBLE_AND_LIMIT_TYPES['bi']['id']
                    level_df.loc[type_filter, 'deductible'] = level_df.loc[type_filter, 'deductible'] * level_df.loc[type_filter, 'agg_tiv'] / 365.

                if 'limit' in level_df.columns and 'lim_type' in level_df.columns:
                    level_df['limit'] = level_df['limit'].fillna(0.)
                    level_df['lim_type'] = level_df['lim_type'].infer_objects(copy=False).fillna(0.).astype('i4')
                    level_df['limit'] = np.where(
                        level_df['lim_type'] == DEDUCTIBLE_AND_LIMIT_TYPES['pctiv']['id'],
                        level_df['limit'] * level_df['agg_tiv'],
                        level_df['limit']
                    )
                    level_df.loc[level_df['lim_type'] == DEDUCTIBLE_AND_LIMIT_TYPES['pctiv']
                                 ['id'], 'lim_type'] = DEDUCTIBLE_AND_LIMIT_TYPES['flat']['id']

                    type_filter = level_df['lim_type'] == DEDUCTIBLE_AND_LIMIT_TYPES['bi']['id']
                    level_df.loc[type_filter, 'limit'] = level_df.loc[type_filter, 'limit'] * level_df.loc[type_filter, 'agg_tiv'] / 365.

                if level_id == SUPPORTED_FM_LEVELS['policy layer']['id']:
                    level_df['calcrule_id'] = get_calc_rule_ids(level_df, calc_rule_type='policy_layer')
                    level_df['profile_id'] = get_profile_ids(level_df) + profile_id_offset
                    profile_id_offset = level_df['profile_id'].max() if not level_df.empty else profile_id_offset
                elif "is_step" in level_df.columns:
                    step_filter = level_df["is_step"]
                    # Before assigning calc. rule IDs and policy TC IDs, the StepTriggerType
                    # should be split into its sub-types in cases where the associated
                    # coverages are covered separately
                    # For example, StepTriggerType = 5 covers buildings and contents separately

                    def assign_sub_step_trigger_type(row):
                        try:
                            step_trigger_type = STEP_TRIGGER_TYPES[row['steptriggertype']]['sub_step_trigger_types'][
                                row['coverage_type_id']]
                            return step_trigger_type
                        except KeyError:
                            return row['steptriggertype']
                    level_df.loc[step_filter, 'steptriggertype'] = level_df[step_filter].apply(
                        lambda row: assign_sub_step_trigger_type(row), axis=1
                    )
                    final_step_filter = level_df['steptriggertype'] > 0
                    # step part
                    level_df.loc[final_step_filter, 'calcrule_id'] = get_calc_rule_ids(level_df[final_step_filter], calc_rule_type='step')
                    has_step = final_step_filter & (~level_df["step_id"].isna())
                    level_df.loc[has_step, 'profile_id'] = factorize_ndarray(level_df.loc[has_step, ['layer_id'] + factorize_key].values,
                                                                             col_idxs=range(len(['layer_id'] + factorize_key)))[0] + profile_id_offset
                    profile_id_offset = level_df.loc[has_step, 'profile_id'].max() if has_step.any() else profile_id_offset

                    # non step part
                    level_df.loc[~final_step_filter, 'calcrule_id'] = get_calc_rule_ids(level_df[~final_step_filter], calc_rule_type='base')
                    level_df.loc[~has_step, 'profile_id'] = (get_profile_ids(level_df[~has_step]) + profile_id_offset)
                    profile_id_offset = level_df.loc[~has_step, 'profile_id'].max() if (~has_step).any() else profile_id_offset
                else:
                    level_df['calcrule_id'] = get_calc_rule_ids(level_df, calc_rule_type='base')
                    level_df['profile_id'] = get_profile_ids(level_df) + profile_id_offset
                    profile_id_offset = level_df['profile_id'].max() if not level_df.empty else profile_id_offset

                write_fm_profile_level(level_df, fm_profile_file, step_policies_present, chunksize=chunksize)

                # =============================================================
                # we have prepared FMTermGroupID on gul or level df (depending on  step_level) now we can merge the terms for this level to gul
                merge_col = list(set(level_df.columns).intersection(set(gul_inputs_df)))
                level_df = level_df[merge_col + ['profile_id']].drop_duplicates()
                # gul_inputs that don't have layer yet must not be merge using layer_id
                non_layered_inputs_df = (
                    gul_inputs_df[gul_inputs_df['layer_id'] == 0]
                    .drop(columns=['layer_id'])
                    .rename(columns={'PolNumber': 'PolNumber_temp'})
                    .merge(level_df, how='left')
                )
                if 'PolNumber' in level_df.columns:
                    # gul_inputs['PolNumber'] is set to level_df['PolNumber'] if empty or if this is the first time layer appear
                    new_layered_gul_input_id = non_layered_inputs_df[non_layered_inputs_df['layer_id'] > 1]['gul_input_id'].unique()
                    non_layered_inputs_df['PolNumber'] = non_layered_inputs_df['PolNumber'].where(
                        (non_layered_inputs_df['gul_input_id'].isin(new_layered_gul_input_id)) | (non_layered_inputs_df['PolNumber_temp'].isna()),
                        non_layered_inputs_df['PolNumber_temp']
                    )
                    non_layered_inputs_df = non_layered_inputs_df.drop(columns=['PolNumber_temp'])
                else:
                    non_layered_inputs_df = non_layered_inputs_df.rename(columns={'PolNumber_temp': 'PolNumber'})

                layered_inputs_df = gul_inputs_df[gul_inputs_df['layer_id'] > 0].merge(level_df, how='left')
                # drop premature layering (no difference of policy between layers)
                if not is_policy_layer_level and level_id not in cross_layer_level:
                    non_layered_inputs_df['layered_id'] = (non_layered_inputs_df['layer_id']
                                                           .where(non_layered_inputs_df['agg_id'].isin(layered_inputs_df['agg_id']), 0))
                    non_layered_inputs_df = (non_layered_inputs_df
                                             .drop_duplicates(subset=['gul_input_id', 'agg_id', 'profile_id', 'layered_id'])
                                             .drop(columns=['layered_id']))

                gul_inputs_df = pd.concat(df for df in [layered_inputs_df, non_layered_inputs_df] if not df.empty)

                gul_inputs_df['layer_id'] = gul_inputs_df['layer_id'].fillna(1).astype('i4')
                gul_inputs_df.sort_values(by=['gul_input_id', 'layer_id'])
                gul_inputs_df["profile_id"] = gul_inputs_df["profile_id"].fillna(1).astype('i4')

                # check rows in prev df that are this level granularity (if prev_agg_id has multiple corresponding agg_id)
                need_root_start_df = gul_inputs_df.groupby("agg_id_prev", observed=True)["agg_id"].nunique()
                need_root_start_df = need_root_start_df[need_root_start_df > 1].index

                gul_inputs_df = gul_inputs_df.drop(columns=["root_start"], errors='ignore')  # clean up
                if need_root_start_df.shape[0]:
                    gul_inputs_df["root_start"] = gul_inputs_df["agg_id"].isin(
                        set(gul_inputs_df.loc[gul_inputs_df["agg_id_prev"].isin(need_root_start_df), 'agg_id'])
                    )

                cur_level_id += 1
                gul_inputs_df['level_id'] = cur_level_id
                # write fm_policytc
                if level_id in cross_layer_level:
                    fm_policytc_df = gul_inputs_df.loc[gul_inputs_df['layer_id'] == 1, fm_policytc_headers]
                else:
                    fm_policytc_df = gul_inputs_df.loc[:, fm_policytc_headers]
                fm_policytc_df.drop_duplicates().astype(fm_policytc_dtype).to_csv(fm_policytc_file, index=False,
                                                                                  header=False, chunksize=chunksize)

                # write programe
                fm_programe_df = gul_inputs_df[['level_id', 'agg_id']].rename(columns={'agg_id': 'to_agg_id'})
                if "root_start" in gul_inputs_df:
                    fm_programe_df['from_agg_id'] = gul_inputs_df['agg_id_prev'].where(~gul_inputs_df["root_start"], -gul_inputs_df['gul_input_id'])
                else:
                    fm_programe_df['from_agg_id'] = gul_inputs_df['agg_id_prev']
                (fm_programe_df[fm_programme_headers].drop_duplicates().astype(fm_programme_dtype)
                 .to_csv(fm_programme_file, index=False, header=False, chunksize=chunksize))

                # reset gul_inputs_df level columns
                gul_inputs_df = (gul_inputs_df
                                 .drop(columns=["root_start", "agg_id_prev", "is_step", "FMTermGroupID",
                                                "profile_id", "level_id", 'fm_peril', 'need_tiv',
                                                'agg_tiv'], errors='ignore')
                                 .rename(columns={"agg_id": "agg_id_prev"}))
                logger.info(f"level {cur_level_id} {level_info} took {time.time()-t0}")
                t0 = time.time()

        gul_inputs_df = gul_inputs_df.sort_values(['gul_input_id', 'layer_id'], kind='stable')
        gul_inputs_df['output_id'] = gul_inputs_df.reset_index().index + 1

        default_pol_agg_key = [v['field'] for v in fm_aggregation_profile[SUPPORTED_FM_LEVELS['policy layer']['id']]['FMAggKey'].values()]
        no_polnumber_df = gul_inputs_df.loc[gul_inputs_df['PolNumber'].isna(), default_pol_agg_key + ['output_id']]
        if not no_polnumber_df.empty:  # empty polnumber, we use accounts_df to set PolNumber based on policy layer agg_key
            gul_inputs_df.loc[gul_inputs_df['PolNumber'].isna(), 'PolNumber'] = no_polnumber_df.merge(
                accounts_df[default_pol_agg_key + ['PolNumber']].drop_duplicates(subset=default_pol_agg_key)
            ).set_index(no_polnumber_df['output_id'] - 1)['PolNumber']

        # write fm_xref
        (gul_inputs_df
         .rename(columns={'gul_input_id': 'agg_id', 'output_id': 'output'})[fm_xref_headers]
         .astype(fm_xref_dtype)
         .to_csv(fm_xref_file, index=False, header=True, chunksize=chunksize))

        # merge acc_idx
        acc_idx_col = list(set(gul_inputs_df.columns).intersection(accounts_df.columns))
        gul_inputs_df_col = list(gul_inputs_df.columns)
        if 'CondTag' in acc_idx_col:
            gul_inputs_df = (
                gul_inputs_df
                .merge(accounts_df[acc_idx_col + ['acc_idx']].drop_duplicates(subset=acc_idx_col),
                       how='left', validate='many_to_one'))
            acc_idx_col.remove('CondTag')
            gul_inputs_df.loc[gul_inputs_df['acc_idx'].isna(), 'acc_idx'] = (gul_inputs_df.loc[gul_inputs_df['acc_idx'].isna(), gul_inputs_df_col].reset_index()
                                                                             .merge(accounts_df[acc_idx_col + ['acc_idx']].drop_duplicates(subset=acc_idx_col),
                                                                                    how='left', validate='many_to_one').set_index('index')['acc_idx'])

        else:
            gul_inputs_df = (
                gul_inputs_df
                .merge(accounts_df[acc_idx_col + ['acc_idx']].drop_duplicates(subset=acc_idx_col),
                       how='left', validate='many_to_one'))

        return gul_inputs_df


def write_empty_policy_layer(gul_inputs_df, cur_level_id, agg_key, fm_policytc_file, fm_programme_file, chunksize):
    base_fm_df = gul_inputs_df[agg_key + ['layer_id', 'agg_id_prev']]
    base_fm_df["agg_id"] = factorize_ndarray(gul_inputs_df.loc[:, agg_key].values, col_idxs=range(len(agg_key)))[0]
    base_fm_df["profile_id"] = 1
    base_fm_df["level_id"] = cur_level_id
    fm_policytc_df = base_fm_df.loc[:, fm_policytc_headers]
    fm_policytc_df.drop_duplicates().astype(fm_policytc_dtype).to_csv(fm_policytc_file, index=False, header=False,
                                                                      chunksize=chunksize)
    fm_programe_df = base_fm_df[['level_id', 'agg_id']].rename(columns={'agg_id': 'to_agg_id'})
    fm_programe_df['from_agg_id'] = base_fm_df['agg_id_prev']
    fm_programe_df[fm_programme_headers].drop_duplicates().astype(fm_programme_dtype).to_csv(fm_programme_file, index=False,
                                                                                             header=False, chunksize=chunksize)


def write_fm_profile_level(level_df, fm_profile_file, step_policies_present, chunksize=100000):
    """
    Writes an FM profile file.

    :param level_df: level_df (fm terms) dataframe
    :type level_df: pandas.DataFrame

    :param fm_profile_file: open file to write to
    :type fm_profile_file: fileObj

    :param step_policies_present: flag to know which type of file to write
    :type step_policies_present: bool

    :param chunksize: chunksize
    :type chunksize: int

    :return: FM profile file path
    :rtype: str
    """
    level_df = level_df.astype({'calcrule_id': 'i4', 'profile_id': 'i4'})
    # Step policies exist
    if step_policies_present:
        fm_profile_df = level_df[list(set(level_df.columns).intersection(set(fm_profile_step_headers + ['steptriggertype'])))].copy()
        for col in fm_profile_step_headers + ['steptriggertype']:
            if col not in fm_profile_df.columns:
                fm_profile_df[col] = 0
        for non_step_name, step_name in profile_cols_map.items():
            if step_name not in fm_profile_df.columns:
                fm_profile_df[step_name] = 0
            fm_profile_df[step_name] = fm_profile_df[step_name].astype(object)
            if non_step_name in level_df.columns:
                fm_profile_df.loc[
                    ~(fm_profile_df['steptriggertype'] > 0), step_name
                ] = level_df.loc[
                    ~(fm_profile_df['steptriggertype'] > 0),
                    non_step_name
                ]
        fm_profile_df.fillna(0, inplace=True)
        fm_profile_df = fm_profile_df.drop_duplicates()

        # Ensure step_id is of int data type and set default value to 1
        fm_profile_df = fm_profile_df.astype(fm_profile_step_dtype)
        fm_profile_df.loc[fm_profile_df['step_id'] == 0, 'step_id'] = 1

        fm_profile_df = fm_profile_df[fm_profile_step_headers].sort_values(by=["profile_id", 'step_id']).drop_duplicates()
    # No step policies
    else:
        # make sure there is no step file in the folder
        fm_profile_df = level_df[list(set(level_df.columns).intersection(set(policytc_cols)))].copy()
        for col in policytc_cols[2:]:
            if col not in fm_profile_df.columns:
                fm_profile_df[col] = 0.

        fm_profile_df = (
            fm_profile_df
            .rename(columns=profile_cols_map)
            .drop_duplicates()
            .assign(share2=0.0, share3=0.0)
            .astype(fm_profile_dtype)[fm_profile_headers]
        )
    try:
        fm_profile_df.to_csv(
            fm_profile_file,
            index=False,
            header=False,
            chunksize=chunksize,
        )
    except (IOError, OSError) as e:
        raise OasisException("Exception raised in 'write_fm_profile_file'", e)



#!/usr/bin/env python

import logging
import numpy as np
from pathlib import Path

from oasislmf.pytools.common.data import resolve_file, write_ndarray_to_fmt_csv, items_dtype
from oasislmf.pytools.common.input_files import read_coverages
from oasislmf.pytools.converters.data import TOOL_INFO
from oasislmf.pytools.gul.common import coverage_type
from oasislmf.pytools.gul.manager import generate_item_map, gul_get_items, read_getmodel_stream

logger = logging.getLogger(__name__)


def get_cdf_data(event_id, damagecdfrecs, recs, rec_idx_ptr, cdf_dtype):
    """Get the cdf data produced by getmodel.
    Note that the input arrays are lists of cdf entries, namely
    the shape on axis=0 is the number of entries.
    Args:
        event_id (int): event_id
        damagecdfrecs (ndarray[damagecdfrec]): cdf record keys
        recs (ndarray[ProbMean]): cdf record values
        rec_idx_ptr (ndarray[int]): array with the indices of `rec` where each cdf record starts.
        cdf_dtype (np.dtype): cdf numpy dtype.
    Returns:
        data (ndarray[cdf_dtype]): cdf data extracted from recs/getmodel.
    """
    assert len(damagecdfrecs) == len(rec_idx_ptr) - 1, "Number of cdfrecs groups does not match number of cdf keys found"

    data = np.zeros(len(recs), dtype=cdf_dtype)
    Nbins = len(rec_idx_ptr) - 1
    idx = 0
    for group_idx in range(Nbins):
        areaperil_id, vulnerability_id = damagecdfrecs[group_idx]
        for bin_index, rec in enumerate(recs[rec_idx_ptr[group_idx]:rec_idx_ptr[group_idx + 1]]):
            data[idx]["event_id"] = event_id
            data[idx]["areaperil_id"] = areaperil_id
            data[idx]["vulnerability_id"] = vulnerability_id
            data[idx]["bin_index"] = bin_index + 1
            data[idx]["prob_to"] = rec["prob_to"]
            data[idx]["bin_mean"] = rec["bin_mean"]
            idx += 1
    return data


def cdf_tocsv(stack, file_in, file_out, file_type, noheader, run_dir):
    headers = TOOL_INFO[file_type]["headers"]
    dtype = TOOL_INFO[file_type]["dtype"]
    fmt = TOOL_INFO[file_type]["fmt"]

    input_path = Path(run_dir, 'input')
    file_in = resolve_file(file_in, "rb", stack)

    if not noheader:
        file_out.write(",".join(headers) + "\n")

    items = gul_get_items(input_path)
    coverages_tiv = read_coverages(input_path)
    coverages = np.zeros(coverages_tiv.shape[0] + 1, coverage_type)
    coverages[1:]['tiv'] = coverages_tiv
    item_map = generate_item_map(items, coverages)
    del coverages_tiv

    compute = np.zeros(coverages.shape[0] + 1, items.dtype['coverage_id'])
    seeds = np.zeros(len(np.unique(items['group_id'])), dtype=items_dtype['group_id'])
    valid_area_peril_id = None

    for event_data in read_getmodel_stream(file_in, item_map, coverages, compute, seeds, valid_area_peril_id):
        event_id, compute_i, items_data, damagecdfrecs, recs, rec_idx_ptr, rng_index = event_data
        data = get_cdf_data(event_id, damagecdfrecs, recs, rec_idx_ptr, dtype)
        write_ndarray_to_fmt_csv(file_out, data, headers, fmt)



from oasis_data_manager.errors import OasisException

__all__ = [
    'OasisException'
]


class OasisStreamException(OasisException):
    def __init__(self, msg, original_exception=None):
        super().__init__(msg, original_exception)



# elt/manager.py

import logging
import numpy as np
import numba as nb
from contextlib import ExitStack
from pathlib import Path
import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq

from oasislmf.pytools.common.data import (DEFAULT_BUFFER_SIZE, MEAN_TYPE_ANALYTICAL, MEAN_TYPE_SAMPLE, oasis_int, oasis_float,
                                          oasis_int_size, oasis_float_size, write_ndarray_to_fmt_csv)
from oasislmf.pytools.common.event_stream import (MAX_LOSS_IDX, MEAN_IDX, EventReader, init_streams_in,
                                                  mv_read, SUMMARY_STREAM_ID)
from oasislmf.pytools.common.input_files import read_event_rates, read_quantile
from oasislmf.pytools.elt.data import MELT_dtype, MELT_fmt, MELT_headers, QELT_dtype, QELT_fmt, QELT_headers, SELT_dtype, SELT_fmt, SELT_headers
from oasislmf.pytools.utils import redirect_logging

logger = logging.getLogger(__name__)


class ELTReader(EventReader):
    def __init__(self, len_sample, compute_selt, compute_melt, compute_qelt, unique_event_ids, event_rates, intervals):
        self.logger = logger

        # Buffer for SELT data
        self.selt_data = np.zeros(DEFAULT_BUFFER_SIZE, dtype=SELT_dtype)
        self.selt_idx = np.zeros(1, dtype=np.int64)

        # Buffer for MELT data
        self.melt_data = np.zeros(DEFAULT_BUFFER_SIZE, dtype=MELT_dtype)
        self.melt_idx = np.zeros(1, dtype=np.int64)

        # Buffer for QELT data
        self.qelt_data = np.zeros(DEFAULT_BUFFER_SIZE, dtype=QELT_dtype)
        self.qelt_idx = np.zeros(1, dtype=np.int64)

        read_buffer_state_dtype = np.dtype([
            ('len_sample', oasis_int),
            ('read_summary_set_id', np.bool_),
            ('reading_losses', np.bool_),
            ('compute_selt', np.bool_),
            ('compute_melt', np.bool_),
            ('compute_qelt', np.bool_),
            ('summary_id', oasis_int),
            ('impacted_exposure', oasis_float),
            ('sumloss', np.float64),  # intermediary value need higher precision to not accumulate error
            ('sumlosssqr', np.float64),  # intermediary value need higher precision to not accumulate error
            ('non_zero_samples', oasis_int),
            ('max_loss', oasis_float),
            ('analytical_mean', oasis_float),
            ('losses_vec', oasis_float, (len_sample,)),
        ])

        self.state = np.zeros(1, dtype=read_buffer_state_dtype)[0]
        self.state["len_sample"] = len_sample
        self.state["reading_losses"] = False
        self.state["read_summary_set_id"] = False
        self.state["compute_selt"] = compute_selt
        self.state["compute_melt"] = compute_melt
        self.state["compute_qelt"] = compute_qelt
        self.unique_event_ids = unique_event_ids.astype(oasis_int)
        self.event_rates = event_rates.astype(oasis_float)
        self.intervals = intervals

        self.curr_file_idx = None  # Current summary file idx being read

    def get_data(self, out_type):
        if out_type == "selt":
            return self.selt_data
        elif out_type == "melt":
            return self.melt_data
        elif out_type == "qelt":
            return self.qelt_data
        else:
            raise RuntimeError(f"Unknown out_type {out_type}")

    def get_data_idx(self, out_type):
        if out_type == "selt":
            return self.selt_idx
        elif out_type == "melt":
            return self.melt_idx
        elif out_type == "qelt":
            return self.qelt_idx
        else:
            raise RuntimeError(f"Unknown out_type {out_type}")

    def read_buffer(self, byte_mv, cursor, valid_buff, event_id, item_id, file_idx):
        # Check for new file idx to read summary_set_id at the start of each summary file stream
        # This is not done by init_streams_in as the summary_set_id is unique to the summary_stream only
        if self.curr_file_idx is not None and self.curr_file_idx != file_idx:
            self.curr_file_idx = file_idx
            self.state["read_summary_set_id"] = False
        else:
            self.curr_file_idx = file_idx

        # Pass state variables to read_buffer
        cursor, event_id, item_id, ret = read_buffer(
            byte_mv, cursor, valid_buff, event_id, item_id, self.selt_data, self.selt_idx,
            self.state, self.melt_data, self.melt_idx, self.qelt_data, self.qelt_idx, self.intervals,
            self.unique_event_ids, self.event_rates
        )
        return cursor, event_id, item_id, ret


@nb.njit(cache=True)
def _update_selt_data(
    selt_data, si, event_id, summary_id,
    sample_id,
    loss,
    impacted_exposure,
):
    selt_data[si]['EventId'] = event_id
    selt_data[si]['SummaryId'] = summary_id
    selt_data[si]['SampleId'] = sample_id
    selt_data[si]['Loss'] = loss
    selt_data[si]['ImpactedExposure'] = impacted_exposure


@nb.njit(cache=True)
def _update_melt_data(
    melt_data, mi, event_id, summary_id,
    sample_type,
    event_rate,
    chance_of_loss,
    meanloss,
    sdloss,
    maxloss,
    footprint_exposure,
    mean_impacted_exposure,
    max_impacted_exposure
):
    melt_data[mi]['EventId'] = event_id
    melt_data[mi]['SummaryId'] = summary_id
    melt_data[mi]['SampleType'] = sample_type
    melt_data[mi]['EventRate'] = event_rate
    melt_data[mi]['ChanceOfLoss'] = chance_of_loss
    melt_data[mi]['MeanLoss'] = meanloss
    melt_data[mi]['SDLoss'] = sdloss
    melt_data[mi]['MaxLoss'] = maxloss
    melt_data[mi]["FootprintExposure"] = footprint_exposure
    melt_data[mi]["MeanImpactedExposure"] = mean_impacted_exposure
    melt_data[mi]["MaxImpactedExposure"] = max_impacted_exposure


@nb.njit(cache=True)
def _update_qelt_data(
    qelt_data, qi, event_id, summary_id,
    quantile,
    loss,
):
    qelt_data[qi]["EventId"] = event_id
    qelt_data[qi]["SummaryId"] = summary_id
    qelt_data[qi]["Quantile"] = quantile
    qelt_data[qi]["Loss"] = loss


@nb.njit(cache=True, error_model="numpy")
def read_buffer(
        byte_mv, cursor, valid_buff, event_id, item_id,
        selt_data, selt_idx, state,
        melt_data, melt_idx,
        qelt_data, qelt_idx, intervals,
        unique_event_ids, event_rates
):
    # Initialise idxs
    last_event_id = event_id
    si = selt_idx[0]
    mi = melt_idx[0]
    qi = qelt_idx[0]

    # Helper functions
    def _update_idxs():
        selt_idx[0] = si
        melt_idx[0] = mi
        qelt_idx[0] = qi

    def _reset_state():
        state["reading_losses"] = False
        state["non_zero_samples"] = 0
        state["max_loss"] = 0
        state["analytical_mean"] = 0
        state["losses_vec"].fill(0)
        state["sumloss"] = 0
        state["sumlosssqr"] = 0

    def _get_mean_and_sd_loss():
        meanloss = state["sumloss"] / state["len_sample"]
        if state["non_zero_samples"] > 1:
            variance = (
                state["sumlosssqr"] - (
                    (state["sumloss"] * state["sumloss"]) / state["len_sample"]
                )
            ) / (state["len_sample"] - 1)

            # Tolerance check
            if variance / state["sumlosssqr"] < 1e-7:
                variance = 0
            sdloss = np.sqrt(variance)
        else:
            sdloss = 0
        return meanloss, sdloss

    while cursor < valid_buff:
        if not state["reading_losses"]:
            # Read summary header
            if valid_buff - cursor >= 3 * oasis_int_size + oasis_float_size:
                # Need to read summary_set_id from summary info first
                if not state["read_summary_set_id"]:
                    _, cursor = mv_read(byte_mv, cursor, oasis_int, oasis_int_size)
                    state["read_summary_set_id"] = True
                event_id_new, cursor = mv_read(byte_mv, cursor, oasis_int, oasis_int_size)
                if last_event_id != 0 and event_id_new != last_event_id:
                    # New event, return to process the previous event
                    _update_idxs()
                    return cursor - oasis_int_size, last_event_id, item_id, 1
                event_id = event_id_new
                state["summary_id"], cursor = mv_read(byte_mv, cursor, oasis_int, oasis_int_size)
                state["impacted_exposure"], cursor = mv_read(byte_mv, cursor, oasis_float, oasis_float_size)
                state["reading_losses"] = True
            else:
                break  # Not enough for whole summary header

        if state["reading_losses"]:
            if valid_buff - cursor < oasis_int_size + oasis_float_size:
                break  # Not enough for whole record

            # Read sidx
            sidx, cursor = mv_read(byte_mv, cursor, oasis_int, oasis_int_size)
            if sidx == 0:  # sidx == 0, end of record
                cursor += oasis_float_size  # Read extra 0 for end of record
                if state["compute_melt"]:
                    # Compute MELT statistics and store them
                    if state["non_zero_samples"] > 0:
                        meanloss, sdloss = _get_mean_and_sd_loss()
                        chance_of_loss = state["non_zero_samples"] / state["len_sample"]
                        mean_imp_exp = state["impacted_exposure"] * chance_of_loss

                        if unique_event_ids.size > 0:
                            idx_ev = np.searchsorted(unique_event_ids, event_id)
                            if idx_ev < unique_event_ids.size and unique_event_ids[idx_ev] == event_id:
                                event_rate = event_rates[idx_ev]
                            else:
                                event_rate = np.nan
                        else:
                            event_rate = np.nan

                        # Update MELT data (analytical mean)
                        _update_melt_data(
                            melt_data, mi,
                            event_id=event_id,
                            summary_id=state["summary_id"],
                            sample_type=MEAN_TYPE_ANALYTICAL,
                            event_rate=event_rate,
                            chance_of_loss=0,
                            meanloss=state["analytical_mean"],
                            sdloss=0,
                            maxloss=state["max_loss"],
                            footprint_exposure=state["impacted_exposure"],
                            mean_impacted_exposure=state["impacted_exposure"],
                            max_impacted_exposure=state["impacted_exposure"]
                        )
                        mi += 1

                        # Update MELT data (sample mean)
                        _update_melt_data(
                            melt_data, mi,
                            event_id=event_id,
                            summary_id=state["summary_id"],
                            sample_type=MEAN_TYPE_SAMPLE,
                            event_rate=event_rate,
                            chance_of_loss=chance_of_loss,
                            meanloss=meanloss,
                            sdloss=sdloss,
                            maxloss=state["max_loss"],
                            footprint_exposure=state["impacted_exposure"],
                            mean_impacted_exposure=mean_imp_exp,
                            max_impacted_exposure=state["impacted_exposure"]
                        )
                        mi += 1

                        if mi >= melt_data.shape[0]:
                            # Output array is full
                            _update_idxs()
                            return cursor, event_id, item_id, 1

                # Update QELT data
                if state["compute_qelt"]:
                    state["losses_vec"].sort()

                    # Calculate loss for per quantile interval
                    for i in range(len(intervals)):
                        q = intervals[i]["quantile"]
                        ipart = intervals[i]["integer_part"]
                        fpart = intervals[i]["fractional_part"]
                        if ipart == len(state["losses_vec"]):
                            loss = state["losses_vec"][ipart - 1]
                        else:
                            loss = (state["losses_vec"][ipart] - state["losses_vec"][ipart - 1]) * fpart + state["losses_vec"][ipart - 1]

                        _update_qelt_data(
                            qelt_data, qi,
                            event_id=event_id,
                            summary_id=state['summary_id'],
                            quantile=q,
                            loss=loss
                        )
                        qi += 1
                        if qi >= qelt_data.shape[0]:
                            # Output array is full
                            _update_idxs()
                            return cursor, event_id, item_id, 1

                # Reset variables
                _reset_state()
                continue

            # Read loss
            loss, cursor = mv_read(byte_mv, cursor, oasis_float, oasis_float_size)
            if sidx == MAX_LOSS_IDX:
                state["max_loss"] = loss
            else:  # Normal data record
                # Update SELT data
                if state["compute_selt"]:
                    _update_selt_data(
                        selt_data, si,
                        event_id=event_id,
                        summary_id=state["summary_id"],
                        sample_id=sidx,
                        loss=loss,
                        impacted_exposure=state["impacted_exposure"]
                    )
                    si += 1
                    if si >= selt_data.shape[0]:
                        # Output array is full
                        _update_idxs()
                        return cursor, event_id, item_id, 1

                if sidx > 0:
                    # Update MELT variables
                    state["sumloss"] += loss
                    state["sumlosssqr"] += loss * loss
                    if loss > 0:
                        state["non_zero_samples"] += 1
                    # Update QELT variables
                    state["losses_vec"][sidx - 1] = loss
            if sidx == MEAN_IDX:
                # Update MELT variables
                state["analytical_mean"] = loss
        else:
            pass  # Should never reach here anyways

    # Update the indices
    _update_idxs()
    return cursor, event_id, item_id, 0


def read_input_files(run_dir, compute_melt, compute_qelt, sample_size):
    """Reads all input files and returns a dict of relevant data
    Args:
        run_dir (str | os.PathLike): Path to directory containing required files structure
        compute_melt (bool): Compute MELT bool
        compute_qelt (bool): Compute QELT bool
        sample_size (int): Sample size
    Returns:
        file_data (Dict[str, Any]): A dict of relevent data extracted from files
    """
    unique_event_ids = np.array([], dtype=oasis_int)
    event_rates = np.array([], dtype=oasis_float)
    include_event_rate = False
    if compute_melt:
        unique_event_ids, event_rates = read_event_rates(Path(run_dir, "input"))
        include_event_rate = unique_event_ids.size > 0

    intervals = read_quantile(sample_size, Path(run_dir, "input"), return_empty=not compute_qelt)

    file_data = {
        "unique_event_ids": unique_event_ids,
        "event_rates": event_rates,
        "include_event_rate": include_event_rate,
        "intervals": intervals,
    }
    return file_data


def run(
    run_dir,
    files_in,
    selt_output_file=None,
    melt_output_file=None,
    qelt_output_file=None,
    noheader=False,
    output_format="csv",
):
    """Runs ELT calculations
    Args:
        run_dir (str | os.PathLike): Path to directory containing required files structure
        files_in (list[str]): Path to summary binary input file
        selt_output_file (str, optional): Path to SELT output file. Defaults to None.
        melt_output_file (str, optional): Path to MELT output file. Defaults to None.
        qelt_output_file (str, optional): Path to QELT output file. Defaults to None.
        noheader (bool): Boolean value to skip header in output file. Defaults to False.
        output_format (str): Output format extension. Defaults to "csv".
    """
    outmap = {
        "selt": {
            "compute": selt_output_file is not None,
            "file_path": selt_output_file,
            "fmt": SELT_fmt,
            "headers": SELT_headers,
            "file": None,
        },
        "melt": {
            "compute": melt_output_file is not None,
            "file_path": melt_output_file,
            "fmt": MELT_fmt,
            "headers": MELT_headers,
            "file": None,
        },
        "qelt": {
            "compute": qelt_output_file is not None,
            "file_path": qelt_output_file,
            "fmt": QELT_fmt,
            "headers": QELT_headers,
            "file": None,
        },
    }

    output_format = "." + output_format
    output_binary = output_format == ".bin"
    output_parquet = output_format == ".parquet"
    # Check for correct suffix
    for path in [v["file_path"] for v in outmap.values()]:
        if path is None:
            continue
        if Path(path).suffix == "":  # Ignore suffix for pipes
            continue
        if (Path(path).suffix != output_format):
            raise ValueError(f"Invalid file extension for {output_format}, got {path},")

    if run_dir is None:
        run_dir = './work'

    if not all([v["compute"] for v in outmap.values()]):
        logger.warning("No output files specified")

    with ExitStack() as stack:
        if files_in == ["-"]:
            files_in = None  # init_streams checks for None to read from sys.stdin.buffer

        streams_in, (stream_source_type, stream_agg_type, len_sample) = init_streams_in(files_in, stack)
        if stream_source_type != SUMMARY_STREAM_ID:
            raise Exception(f"unsupported stream type {stream_source_type}, {stream_agg_type}")

        file_data = read_input_files(
            run_dir,
            outmap["melt"]["compute"],
            outmap["qelt"]["compute"],
            len_sample
        )
        elt_reader = ELTReader(
            len_sample,
            outmap["selt"]["compute"],
            outmap["melt"]["compute"],
            outmap["qelt"]["compute"],
            file_data["unique_event_ids"],
            file_data["event_rates"],
            file_data["intervals"]
        )

        # Initialise output files ELT
        if output_binary:
            for out_type in outmap:
                if not outmap[out_type]["compute"]:
                    continue
                out_file = stack.enter_context(open(outmap[out_type]["file_path"], 'wb'))
                outmap[out_type]["file"] = out_file
        elif output_parquet:
            for out_type in outmap:
                if not outmap[out_type]["compute"]:
                    continue
                temp_df = pd.DataFrame(elt_reader.get_data(out_type), columns=outmap[out_type]["headers"])
                temp_table = pa.Table.from_pandas(temp_df)
                out_file = pq.ParquetWriter(outmap[out_type]["file_path"], temp_table.schema)
                outmap[out_type]["file"] = out_file
        else:
            for out_type in outmap:
                if not outmap[out_type]["compute"]:
                    continue
                out_file = stack.enter_context(open(outmap[out_type]["file_path"], 'w'))
                if not noheader:
                    csv_headers = ','.join(outmap[out_type]["headers"])
                    out_file.write(csv_headers + '\n')
                outmap[out_type]["file"] = out_file

        # Process summary files
        for event_id in elt_reader.read_streams(streams_in):
            for out_type in outmap:
                if not outmap[out_type]["compute"]:
                    continue

                data_idx = elt_reader.get_data_idx(out_type)
                data = elt_reader.get_data(out_type)[:data_idx[0]]

                if outmap[out_type]["file"] is not None and data.size > 0:
                    if output_binary:
                        data.tofile(outmap[out_type]["file"])
                    elif output_parquet:
                        data_df = pd.DataFrame(data)
                        data_table = pa.Table.from_pandas(data_df)
                        outmap[out_type]["file"].write_table(data_table)
                    else:
                        write_ndarray_to_fmt_csv(
                            outmap[out_type]["file"],
                            data,
                            outmap[out_type]["headers"],
                            outmap[out_type]["fmt"]
                        )
                data_idx[0] = 0


@redirect_logging(exec_name='eltpy')
def main(run_dir='.', files_in=None, selt=None, melt=None, qelt=None, noheader=False, ext="csv", **kwargs):
    run(
        run_dir,
        files_in,
        selt_output_file=selt,
        melt_output_file=melt,
        qelt_output_file=qelt,
        noheader=noheader,
        output_format=ext,
    )



import csv
import json
import logging
from pathlib import Path

from oasislmf.utils.documentation.jsontomd.base import BaseJsonToMarkdownGenerator

logger = logging.getLogger(__name__)


class RDLS_0_2_0_JsonToMarkdownGenerator(BaseJsonToMarkdownGenerator):
    def json_array_to_mdtable(self, data, ref):
        array_schema = self._resolve_internal_ref(ref)
        array_keys = array_schema["properties"].keys()
        headers = []
        for k in array_keys:
            if "title" in array_schema["properties"][k]:
                headers.append(array_schema["properties"][k]["title"])
            else:
                headers.append(k)
        rows = []
        for entry in data:
            row = []
            for k in array_keys:
                v = entry.get(k, "")
                if isinstance(v, list):
                    rets = []
                    for v_ in v:
                        if isinstance(v_, dict):
                            v_ = json.dumps(v_, indent=4).replace('\n', '<br>').replace(' ', '&nbsp;')
                        rets.append(str(v_))
                    v = ",<br>".join(rets)
                elif isinstance(v, dict):
                    pretty_json = json.dumps(v, indent=4).replace('\n', '<br>').replace(' ', '&nbsp;')
                    v = f"<details><summary>Expand</summary>{pretty_json}</details>"
                else:
                    v = str(v)
                row.append(v)
            rows.append(row)
        self.md.add_table(headers, rows)

    def _remove_duplicate_dicts_by_field(self, dicts, field):
        seen_fields = set()
        unique_dicts = []
        for d in dicts:
            if d[field] not in seen_fields:
                unique_dicts.append(d)
                seen_fields.add(d[field])
        sorted_unique_dicts = sorted(unique_dicts, key=lambda d: d[field])
        return sorted_unique_dicts

    def generate_ds_overview(self, data, properties_schema, header_level):
        ds_section_properties = [
            "id",
            "title",
            "description",
            "risk_data_type",
            "version",
            "purpose",
            "project",
            "details",
        ]

        self.md.add_header("Dataset Overview", level=header_level)

        present_properties = [p for p in ds_section_properties if p in data]
        for p in present_properties:
            if p == "risk_data_type":
                self.md.add_definition(properties_schema[p]["title"], ",".join(data[p]))
            else:
                self.md.add_definition(properties_schema[p]["title"], data[p])

        return set(ds_section_properties)

    def _gen_ds_hazard(self, data, properties_schema, header_level):
        event_set_data = data["hazard"]["event_sets"]

        self.md.add_header(properties_schema["hazard"]["title"], level=header_level)
        items_ref = properties_schema["hazard"]["properties"]["event_sets"]["items"]["$ref"]
        ref_properties_schema = self._resolve_internal_ref(items_ref)["properties"]
        period_ref = ref_properties_schema["temporal"]["$ref"]
        period_properties_schema = self._resolve_internal_ref(period_ref)["properties"]
        spatial_id = 0
        spatial_data = []
        hazard_items = []
        events_items = []
        for es_item in event_set_data:
            hazard_items += es_item["hazards"]
            es_item["hazards"] = [hz["id"] for hz in es_item["hazards"]]
            if "spatial" in es_item:
                spatial_data.append(es_item["spatial"])
                es_item["spatial"] = f"loc_{spatial_id}"
                spatial_id += 1
            if "temporal" in es_item:
                es_item["temporal"] = [period_properties_schema[k]["title"] + ": " + str(v) for k, v in es_item["temporal"].items()]
            if "events" in es_item:
                events_items += es_item["events"]
                es_item["events"] = ", ".join([ev["id"] for ev in es_item["events"]])
        self.json_array_to_mdtable(event_set_data, items_ref)

        if spatial_data:
            self.md.add_header(ref_properties_schema["spatial"]["title"], level=header_level + 1)
            spatial_ref = ref_properties_schema["spatial"]["$ref"]
            spatial_properties_schema = self._resolve_internal_ref(spatial_ref)["properties"]
            for i, sp_data in enumerate(spatial_data):
                self.md.add_header(f"loc_{i}", level=header_level + 2)
                self._gen_ds_spatial_data(sp_data, spatial_properties_schema, header_level + 2)

        if events_items:
            self.md.add_header(ref_properties_schema["events"]["title"], level=header_level + 1)
            events_items = self._remove_duplicate_dicts_by_field(events_items, "id")
            events_items_ref = ref_properties_schema["events"]["items"]["$ref"]
            events_items_properties_schema = self._resolve_internal_ref(events_items_ref)["properties"]
            disaster_identifiers_items = []
            footprints_items = []
            for ev_item in events_items:
                hazard_items.append(ev_item["hazard"])
                ev_item["hazard"] = ev_item["hazard"]["id"]
                disaster_identifiers_items += ev_item["disaster_identifiers"]
                ev_item["disaster_identifiers"] = [di["id"] for di in ev_item["disaster_identifiers"]]
                if "footprints" in ev_item:
                    footprints_items += ev_item["footprints"]
                    ev_item["footprints"] = [fp["id"] for fp in ev_item["footprints"]]
            self.json_array_to_mdtable(events_items, events_items_ref)

            self.md.add_header(events_items_properties_schema["disaster_identifiers"]["title"], level=header_level + 2)
            disaster_identifiers_items = self._remove_duplicate_dicts_by_field(disaster_identifiers_items, "id")
            class_items_ref = events_items_properties_schema["disaster_identifiers"]["items"]["$ref"]
            self.json_array_to_mdtable(disaster_identifiers_items, class_items_ref)

            self.md.add_header(events_items_properties_schema["footprints"]["title"], level=header_level + 2)
            footprints_items = self._remove_duplicate_dicts_by_field(footprints_items, "id")
            footprints_items_ref = events_items_properties_schema["footprints"]["items"]["$ref"]
            self.json_array_to_mdtable(footprints_items, footprints_items_ref)

        self.md.add_header(ref_properties_schema["hazards"]["title"], level=header_level + 1)
        hazard_items = self._remove_duplicate_dicts_by_field(hazard_items, "id")
        hazard_items_ref = ref_properties_schema["hazards"]["items"]["$ref"]
        hazard_items_properties_schema = self._resolve_internal_ref(hazard_items_ref)["properties"]
        trigger_ref = hazard_items_properties_schema["trigger"]["$ref"]
        trigger_ref_properties_schema = self._resolve_internal_ref(trigger_ref)["properties"]
        for hz_item in hazard_items:
            if "trigger" in hz_item:
                hz_item["trigger"] = [trigger_ref_properties_schema[k]["title"] + ": " + str(v) for k, v in hz_item["trigger"].items()]
        self.json_array_to_mdtable(hazard_items, hazard_items_ref)

    def _gen_ds_exposure(self, data, properties_schema, header_level):
        exposure_data = data["exposure"]

        self.md.add_header(properties_schema["exposure"]["title"], level=header_level)
        self.md.add_definition(properties_schema["exposure"]["properties"]["category"]["title"], exposure_data["category"])
        if "taxonomy" in exposure_data:
            self.md.add_definition(properties_schema["exposure"]["properties"]["taxonomy"]["title"], exposure_data["taxonomy"])

        exposure_properties_schema = properties_schema["exposure"]["properties"]
        self.md.add_header(exposure_properties_schema["metrics"]["title"], level=header_level + 1)
        gazetteer_items_ref = exposure_properties_schema["metrics"]["items"]["$ref"]
        self.json_array_to_mdtable(exposure_data["metrics"], gazetteer_items_ref)

    def _gen_ds_vulnerability(self, data, properties_schema, header_level):
        vuln_data = data["vulnerability"]
        self.md.add_header(properties_schema["vulnerability"]["title"], level=header_level)

        vuln_properties_schema = properties_schema["vulnerability"]["properties"]

        self.md.add_header("Hazard info", level=header_level + 1)
        headers = ["Key", "Value"]
        rows = []
        rows.append([vuln_properties_schema["hazard_primary"]["title"], vuln_data["hazard_primary"]])
        if "hazard_secondary" in vuln_data:
            rows.append([vuln_properties_schema["hazard_secondary"]["title"], vuln_data["hazard_secondary"]])
        if "hazard_process_primary" in vuln_data:
            rows.append([vuln_properties_schema["hazard_process_primary"]["title"], vuln_data["hazard_process_primary"]])
        if "hazard_process_secondary" in vuln_data:
            rows.append([vuln_properties_schema["hazard_process_secondary"]["title"], vuln_data["hazard_process_secondary"]])
        if "hazard_analysis_type" in vuln_data:
            rows.append([vuln_properties_schema["hazard_analysis_type"]["title"], vuln_data["hazard_analysis_type"]])
        rows.append([vuln_properties_schema["intensity"]["title"], vuln_data["intensity"]])
        self.md.add_table(headers, rows)

        self.md.add_header("Exposure info", level=header_level + 1)
        self.md.add_definition(vuln_properties_schema["category"]["title"], vuln_data["category"])
        if "taxonomy" in vuln_data:
            self.md.add_definition(vuln_properties_schema["taxonomy"]["title"], vuln_data["taxonomy"])
        cost_items_ref = vuln_properties_schema["cost"]["items"]["$ref"]
        self.json_array_to_mdtable(vuln_data["cost"], cost_items_ref)

        self.md.add_header("Vulnerability Impact info", level=header_level + 1)
        impact_ref = vuln_properties_schema["impact"]["$ref"]
        impact_properties_schema = self._resolve_internal_ref(impact_ref)["properties"]
        impact_data = vuln_data["impact"]
        if "type" in impact_data:
            self.md.add_definition(impact_properties_schema["type"]["title"], impact_data["type"])
        if "metric" in impact_data:
            self.md.add_definition(impact_properties_schema["metric"]["title"], impact_data["metric"])
        if "unit" in impact_data:
            self.md.add_definition(impact_properties_schema["unit"]["title"], impact_data["unit"])
        if "base_data_type" in impact_data:
            self.md.add_definition(impact_properties_schema["base_data_type"]["title"], impact_data["base_data_type"])

        self.md.add_header("Vulnerability Spatial info", level=header_level + 1)
        spatial_ref = vuln_properties_schema["spatial"]["$ref"]
        spatial_properties_schema = self._resolve_internal_ref(spatial_ref)["properties"]
        spatial_data = vuln_data["spatial"]
        self._gen_ds_spatial_data(spatial_data, spatial_properties_schema, header_level)

        self.md.add_header("Vulnerability Functions info", level=header_level + 1)
        functions_data = vuln_data["functions"]
        funcs_properties_schema = vuln_properties_schema["functions"]["properties"]
        for f_id, f_data in functions_data.items():
            self.md.add_header(funcs_properties_schema[f_id]["title"], level=header_level + 2)
            headers = ["Key", "Value"]
            rows = [[funcs_properties_schema[f_id]["properties"][k]["title"], v] for k, v in f_data.items()]
            self.md.add_table(headers, rows)

        if "analysis_details" in vuln_data or "se_category" in vuln_data:
            self.md.add_header("Vulnerability Extra info", level=header_level + 1)
            if "analysis_details" in vuln_data:
                self.md.add_definition(vuln_properties_schema["analysis_details"]["title"], vuln_data["analysis_details"])
            if "se_category" in vuln_data:
                self.md.add_header(vuln_properties_schema["se_category"]["title"], header_level + 2)
                class_ref = vuln_properties_schema["se_category"]["$ref"]
                class_properties_schema = self._resolve_internal_ref(class_ref)["properties"]
                headers = ["Key", "Value"]
                rows = [[class_properties_schema[k]["title"], v] for k, v in vuln_data["se_category"].items()]
                self.md.add_table(headers, rows)

    def _gen_ds_loss(self, data, properties_schema, header_level):
        loss_data = data["loss"]["losses"]

        self.md.add_header(properties_schema["loss"]["title"], level=header_level)
        items_ref = properties_schema["loss"]["properties"]["losses"]["items"]["$ref"]
        ref_properties_schema = self._resolve_internal_ref(items_ref)["properties"]

        cost_items = []
        for loss_item in loss_data:
            cost_items.append(loss_item["cost"])
            loss_item["cost"] = loss_item["cost"]["id"]
            if "impact" in loss_item:
                impact_items_ref = ref_properties_schema["impact"]["$ref"]
                impact_ref_properties_schema = self._resolve_internal_ref(impact_items_ref)["properties"]
                loss_item["impact"] = [impact_ref_properties_schema[k]["title"] + ": " + v for k, v in loss_item["impact"].items()]

        self.json_array_to_mdtable(loss_data, items_ref)

        cost_items = self._remove_duplicate_dicts_by_field(cost_items, "id")
        cost_items_ref = ref_properties_schema["cost"]["$ref"]
        self.md.add_header(ref_properties_schema["cost"]["title"], level=header_level + 1)
        self.json_array_to_mdtable(cost_items, cost_items_ref)

    def generate_ds_risk_data_properties(self, data, properties_schema, header_level):
        ds_section_properties = [
            "hazard",
            "exposure",
            "vulnerability",
            "loss",
        ]
        risk_data_types = data["risk_data_type"]
        present_properties = [p for p in ds_section_properties if p in data]

        missing_properties = set(risk_data_types) - set(present_properties)
        extra_properties = set(present_properties) - set(risk_data_types)
        if len(missing_properties) > 0:
            logger.warning(f"Warning: Missing risk data types from json: {list(missing_properties)}")
        if len(extra_properties) > 0:
            logger.warning(f"Warning: Extra risk data types found in json ({list(extra_properties)}) will also be output")

        if "hazard" in data:
            self._gen_ds_hazard(data, properties_schema, header_level)
        if "exposure" in data:
            self._gen_ds_exposure(data, properties_schema, header_level)
        if "vulnerability" in data:
            self._gen_ds_vulnerability(data, properties_schema, header_level)
        if "loss" in data:
            self._gen_ds_loss(data, properties_schema, header_level)

        return set(ds_section_properties)

    def _gen_ds_spatial_data(self, spatial_data, ref_properties_schema, header_level):
        if "scale" in spatial_data:
            self.md.add_definition(ref_properties_schema["scale"]["title"], spatial_data["scale"])
        if "countries" in spatial_data:
            self.md.add_definition(ref_properties_schema["countries"]["title"], ",".join(spatial_data["countries"]))
        if "centroid" in spatial_data:
            self.md.add_definition(ref_properties_schema["centroid"]["title"], spatial_data["centroid"])
        if "bbox" in spatial_data:
            self.md.add_definition(ref_properties_schema["bbox"]["title"], spatial_data["bbox"])
        if "gazetteer_entries" in spatial_data:
            self.md.add_header(ref_properties_schema["gazetteer_entries"]["title"], level=header_level + 1)
            gazetteer_items_ref = ref_properties_schema["gazetteer_entries"]["items"]["$ref"]
            self.json_array_to_mdtable(spatial_data["gazetteer_entries"], gazetteer_items_ref)
        if "geometry" in spatial_data:
            self.md.add_header(ref_properties_schema["geometry"]["title"], level=header_level + 1)
            geometry_ref_properties_schema = self._resolve_internal_ref(ref_properties_schema["geometry"]["$ref"])["properties"]
            if "type" in geometry_ref_properties_schema:
                self.md.add_definition(geometry_ref_properties_schema["type"]["title"], spatial_data["geometry"]["type"])
            if "coordinates" in geometry_ref_properties_schema:
                self.md.add_collapsible_section(json.dumps(spatial_data["geometry"]["coordinates"], indent=2), "Raw Coordinates")

    def generate_ds_spatial_temporal_properties(self, data, properties_schema, header_level):
        ds_section_properties = [
            "spatial",
            "temporal_resolution",
        ]

        self.md.add_header("Spatial and Temporal Coverage", level=header_level)
        temporal_data = data.get("temporal_resolution", "No temporal information found in json.")
        self.md.add_definition(properties_schema["temporal_resolution"]["title"], temporal_data)

        self.md.add_header("Spatial", level=header_level + 1)
        ref_properties_schema = self._resolve_internal_ref(properties_schema["spatial"]["$ref"])["properties"]
        spatial_data = data["spatial"]
        self._gen_ds_spatial_data(spatial_data, ref_properties_schema, header_level)

        return set(ds_section_properties)

    def generate_ds_resources_properties(self, data, properties_schema, header_level):
        ds_section_properties = [
            "resources",
            "sources",
        ]

        resources_data = data["resources"]

        self.md.add_header("Resources and Sources", level=header_level)
        self.md.add_header(properties_schema["resources"]["title"], level=header_level + 1)
        items_ref = properties_schema["resources"]["items"]["$ref"]
        self.json_array_to_mdtable(resources_data, items_ref)

        for entry in resources_data:
            self.md.add_header(entry.get("title", "Unnamed Resource"), level=header_level + 2)
            fmt = entry.get("format", "unknown")

            if "download_url" not in entry:
                self.md.add_text("No path found to display data")
                continue

            fp = Path(self.data_path, entry["download_url"])
            if not fp.exists():
                self.md.add_text(f"No file found at {fp}, could not display data")
                continue

            self.md.add_text(f"File ({fp.name}) found [here]({fp.relative_to(self.doc_out_dir.resolve()).as_posix()})")

            if fmt == "csv":
                with open(fp) as f:
                    reader = csv.DictReader(f)
                    headers = reader.fieldnames or []
                    rows = [[str(row.get(h, "")) for h in headers] for _, row in zip(range(10), reader)]
                self.md.add_text("First 10 rows displayed only")
                self.md.add_table(headers, rows)
            else:
                self.md.add_text(f"Cannot display preview for {fmt} files")

        if "sources" in data:
            self.md.add_header(properties_schema["sources"]["title"], level=header_level + 1)
            sources_data = data["sources"]
            items_ref = properties_schema["sources"]["items"]["$ref"]
            self.json_array_to_mdtable(sources_data, items_ref)

        return set(ds_section_properties)

    def generate_ds_owner_contact_properties(self, data, properties_schema, header_level):
        ds_section_properties = [
            "publisher",
            "creator",
            "contact_point",
            "attributions",
        ]

        self.md.add_header("Ownership and Contacts", level=header_level)

        present_properties = [p for p in ["publisher", "creator", "contact_point"] if p in data]
        for p in present_properties:
            self.md.add_header(properties_schema[p]["title"], level=header_level + 1)
            ref_properties_schema = self._resolve_internal_ref(properties_schema[p]["$ref"])["properties"]
            headers = [ref_properties_schema[k]["title"] for k in data[p].keys()]
            rows = [data[p].values()]
            self.md.add_table(headers, rows)

        if "attributions" in data:
            self.md.add_header(properties_schema["attributions"]["title"], level=header_level + 1)
            attributions_data = data["attributions"]
            for v in attributions_data:
                v["entity"] = list(v["entity"].values())
            items_ref = properties_schema["attributions"]["items"]["$ref"]
            self.json_array_to_mdtable(attributions_data, items_ref)

        return set(ds_section_properties)

    def generate_ds_licensing_links_properties(self, data, properties_schema, header_level):
        ds_section_properties = [
            "license",
            "links",
            "referenced_by",
        ]

        self.md.add_header("Licensing and Links", level=header_level)
        self.md.add_definition(properties_schema["license"]["title"], data["license"])
        links_data = data.get("links", [])
        if links_data:
            self.md.add_header("Links", level=header_level + 1)
            self.md.add_list([x["href"] for x in data["links"]])

        if "referenced_by" in data:
            self.md.add_header(properties_schema["referenced_by"]["title"], level=header_level + 1)
            items_ref = properties_schema["referenced_by"]["items"]["$ref"]
            self.json_array_to_mdtable(data["referenced_by"], items_ref)

        return set(ds_section_properties)

    def generate_dataset(self, data, properties_schema, header_level):
        all_properties = set(self.full_schema["properties"].keys())
        ds_overview_properties = self.generate_ds_overview(data, properties_schema, header_level)
        ds_risk_data_properties = self.generate_ds_risk_data_properties(data, properties_schema, header_level)
        ds_spatial_temporal_properties = self.generate_ds_spatial_temporal_properties(data, properties_schema, header_level)
        ds_resources_properties = self.generate_ds_resources_properties(data, properties_schema, header_level)
        ds_owner_contact_properties = self.generate_ds_owner_contact_properties(data, properties_schema, header_level)
        ds_licensing_links_properties = self.generate_ds_licensing_links_properties(data, properties_schema, header_level)

        missing_properties = all_properties - (
            ds_overview_properties |
            ds_risk_data_properties |
            ds_spatial_temporal_properties |
            ds_resources_properties |
            ds_owner_contact_properties |
            ds_licensing_links_properties
        )

        if len(missing_properties) > 0:
            logger.warning(f"Warning: The following dataset properties have not been implemented for markdown output: {list(missing_properties)}")

    def generate(self, json_data, generate_toc=False):
        self.md.add_header("Documentation", level=1)
        for dataset in json_data.get("datasets", []):
            self.md.add_header(dataset["title"], level=1)
            self.generate_dataset(dataset, self.full_schema["properties"], header_level=2)
            self.md.add_text("")
        return self.md.get_markdown(generate_toc=generate_toc)



import msgpack
import numpy as np
import struct
from oasislmf.pytools.common.data import resolve_file
from oasislmf.pytools.converters.data import TOOL_INFO
import pandas as pd


def complex_items_tobin(stack, file_in, file_out, file_type):
    header_dtype = TOOL_INFO[file_type]["dtype"]

    file_in = resolve_file(file_in, "r", stack)

    s = struct.Struct('IIII')  # Matches dtype from complex_items_meta_output
    try:
        items_df = pd.read_csv(file_in)
    except pd.errors.EmptyDataError:
        np.empty(0, dtype=header_dtype).tofile(file_out)
        return
    for row in items_df.itertuples():
        # item_id,coverage_id,model_data,group_id
        packed_model_data = msgpack.packb(row.model_data)
        values = (
            int(row.item_id),
            int(row.coverage_id),
            int(float(row.group_id)),
            len(packed_model_data)
        )
        packed_data = s.pack(*values)
        file_out.write(packed_data)
        file_out.write(packed_model_data)



import zlib
import numpy as np
import pandas as pd

from oasislmf.pytools.common.data import resolve_file
from oasislmf.pytools.converters.csvtobin.utils.common import read_csv_as_ndarray
from oasislmf.pytools.converters.data import TOOL_INFO
from oasislmf.pytools.getmodel.common import Event_dtype, EventIndexBin_dtype, EventIndexBinZ_dtype
from oasislmf.utils.exceptions import OasisException


def _validate(data):
    df = pd.DataFrame(data)

    # Check probability sums to 1 for each (event_id, areaperil_id) group
    prob_sums = df.groupby(["event_id", "areaperil_id"])["probability"].sum()
    invalid_sums = prob_sums[~np.isclose(prob_sums, 1, atol=1e-6)]
    if not invalid_sums.empty:
        error_msg = "\n".join([
            f"Group (event_id={idx[0]}, areaperil_id={idx[1]}) has prob sum = {val:.6f}"
            for idx, val in invalid_sums.items()
        ])
        raise OasisException(f"Error: Probabilities do not sum to 1 for the following groups: \n{error_msg}")

    # Check sorted by event_id, areaperil_id
    expected_order = df.sort_values(['event_id', 'areaperil_id']).reset_index(drop=True)
    if not df[['event_id', 'areaperil_id']].equals(expected_order[['event_id', 'areaperil_id']]):
        unordered_rows = df[['event_id', 'areaperil_id']].ne(expected_order[['event_id', 'areaperil_id']]).any(axis=1)
        mismatch_indices = df.index[unordered_rows].tolist()
        raise OasisException(f"IDs not in ascending order. First few mismatched indices: \n{df.iloc[mismatch_indices[:10]]}")

    # Check intensity bin uniqueness for each (event_id, areaperil_id) group
    duplicates = df.duplicated(subset=['event_id', 'areaperil_id', 'intensity_bin_id'], keep=False)
    if duplicates.any():
        dup_rows = df[duplicates]
        error_msg = dup_rows[['event_id', 'areaperil_id', 'intensity_bin_id']].drop_duplicates(keep="last").to_string()
        raise OasisException(f"Error: Duplicate intensity bins found: \n{error_msg}")


def footprint_tobin(
    stack, file_in, file_out, file_type,
    idx_file_out,
    zip_files,
    max_intensity_bin_idx,
    no_intensity_uncertainty,
    decompressed_size,
    no_validation
):
    headers = TOOL_INFO[file_type]["headers"]
    dtype = TOOL_INFO[file_type]["dtype"]
    idx_file_out = resolve_file(idx_file_out, "wb", stack)
    data = read_csv_as_ndarray(stack, file_in, headers, dtype)

    if not no_validation:
        _validate(data)

    # Write bin file header
    np.array([max_intensity_bin_idx], dtype=np.int32).tofile(file_out)
    zip_opts = decompressed_size << 1 | (not no_intensity_uncertainty)
    np.array([zip_opts], dtype=np.int32).tofile(file_out)

    offset = np.dtype(np.int32).itemsize * 2

    unique_events = np.unique(data["event_id"])
    for event_id in unique_events:
        event_mask = data["event_id"] == event_id
        event_data = data[event_mask]

        bin_data = np.empty(len(event_data), dtype=Event_dtype)
        bin_data["areaperil_id"] = event_data["areaperil_id"]
        bin_data["intensity_bin_id"] = event_data["intensity_bin_id"]
        bin_data["probability"] = event_data["probability"]

        if any(bin_data["intensity_bin_id"] > max_intensity_bin_idx):
            raise OasisException(f"Error: Found intensity_bin_idx in data larger than max_intensity_bin_idx: {max_intensity_bin_idx}")

        bin_data = bin_data.tobytes()
        dsize = len(bin_data)
        if zip_files:
            bin_data = zlib.compress(bin_data)

        file_out.write(bin_data)
        size = len(bin_data)
        if decompressed_size:
            np.array([(event_id, offset, size, dsize)], dtype=EventIndexBinZ_dtype).tofile(idx_file_out)
        else:
            np.array([(event_id, offset, size)], dtype=EventIndexBin_dtype).tofile(idx_file_out)
        offset += size



from oasislmf.pytools.elt.data import QELT_headers, QELT_dtype, QELT_fmt


headers = QELT_headers
dtype = QELT_dtype
fmt = QELT_fmt
cli_support = ['bintocsv', 'csvtobin', 'bintoparquet', 'parquettobin']



from oasislmf.pytools.plt.data import QPLT_headers, QPLT_dtype, QPLT_fmt


headers = QPLT_headers
dtype = QPLT_dtype
fmt = QPLT_fmt
cli_support = ['bintocsv', 'csvtobin', 'bintoparquet', 'parquettobin']
